/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

#include "ns3/core-module.h"
#include "ns3/common-module.h"
#include "ns3/node-module.h"
#include "ns3/helper-module.h"
#include "ns3/contrib-module.h"
#include "ns3/wifi-module.h"
#include "ns3/mobility-module.h"
#include "ns3/traffic-application.h"

#include <iostream>
#include <fstream>
#include <numeric>

NS_LOG_COMPONENT_DEFINE("Main");

using namespace ns3;

#include "eventtrace.h"

class QosTrafficPacketFactory : public TrafficPacketFactory
{
public:
  static TypeId GetTypeId();

  QosTrafficPacketFactory();

  /**
   * Creates a packet, applies  and returns a pointer to it.
   *
   * \return pointer to the created packet
   */
  virtual Ptr<Packet> CreatePacket();

protected:
  /// Size of packets generated.
  unsigned int  m_size;
  WifiQosTag::AccessCategory m_ac;
};


TypeId 
QosTrafficPacketFactory::GetTypeId()
{
  static TypeId tid = TypeId("ns3::QosTrafficPacketFactory")
    .SetParent<TrafficPacketFactory>()
    .AddAttribute("Size",
                  "Size in bytes of packets generated.",
                  UintegerValue(100),
                  MakeUintegerAccessor(&QosTrafficPacketFactory::m_size),
                  MakeUintegerChecker<unsigned int>())
    .AddAttribute("AC",
                  "The AccessCategory set on the packet.",
                  EnumValue (WifiQosTag::AC_BE),
                  MakeEnumAccessor(&QosTrafficPacketFactory::m_ac),
                  MakeEnumChecker(WifiQosTag::AC_VO, "AC_VO",
                                  WifiQosTag::AC_VI, "AC_VI",
                                  WifiQosTag::AC_BE, "AC_BE",
                                  WifiQosTag::AC_BK, "AC_BK"))
    ;
  return tid;
}

QosTrafficPacketFactory::QosTrafficPacketFactory()
{}

Ptr<Packet>
QosTrafficPacketFactory::CreatePacket()
{
  Ptr<Packet> packet = Create<Packet>(m_size);
  WifiQosTag qostag;
  qostag.SetAc(m_ac);
  packet->AddTag(qostag);
  return packet;
}

class Experiment
{
private:

  static const double simulationTime = 60.0;

  int           m_nodeNum;

  uint32_t      m_packetsTotalAt0[4];
  uint64_t      m_bytesTotalAt0[4];

public:
  double
  GetThroughput(WifiQosTag::AccessCategory AC) const
  {
    double throughPut = 8 * m_bytesTotalAt0[AC] / simulationTime;

    // std::cout << std::fixed << "throughPut[" << AC << "]: " << throughPut << "\n";

    return throughPut;
  }

private:
  void
  PacketSinkRxTrace(std::string context, Ptr<const Packet> p, const Address &ad)
  {
    WifiQosTag qostag;
    if (!p->FindFirstMatchingTag(qostag))
      {
        NS_ASSERT_MSG(false, "No QoS tag");
        exit(-1);
        return;
      }

    if (0)
      {
        std::cerr << Simulator::Now() << " " << context
                  << " PacketSinkRx to=" << ad << " aci=" << qostag.GetAc() << " size=" << p->GetSize() << " p={" << *p << "}" << std::endl;
      }

    int aci = qostag.GetAc();
    NS_ASSERT_MSG(aci >= 0 && aci < 4, "Invalid AC");

    m_packetsTotalAt0[aci]++;
    m_bytesTotalAt0[aci] += p->GetSize();
  }

  void
  DevTxTrace(std::string context, Ptr<const Packet> p, Mac48Address address)
  {
    std::cerr << context << " DevTx to=" << address << " p={" << p->GetSize() << "}" << std::endl;
  }

  void
  PhyTxTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag)
  {
    std::cerr << Simulator::Now() << " PHYTX phytag={" << *phytag << "} size=" << p->GetSize() << " p={" << *p << "}" << std::endl;
  }

  void
  PhyRxOkTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag)
  {
    std::cerr << Simulator::Now() << " PHYRXOK phytag={" << *phytag << "} p={" << *p << "}" << std::endl;
  }

  void
  PhyRxErrorTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag, WifiPhy::RxErrorReason rxErrorReason)
  {
    std::cerr << Simulator::Now() << " PHYRXERROR"
              << " reason=" << WifiPhy::RxErrorReasonToString(rxErrorReason)
              << " phytag={" << *phytag << "}"
              << " packet={" << *p << "}" << std::endl;
  }

public:

  void
  RunOne(int nodeNum, const std::string& edcaParameterSet, bool unicast)
  {
    m_nodeNum = nodeNum;

    memset(m_packetsTotalAt0, 0, sizeof(m_packetsTotalAt0));
    memset(m_bytesTotalAt0, 0, sizeof(m_bytesTotalAt0));

    NodeContainer nodeListener;
    nodeListener.Create(1);

    NodeContainer nodeSender;
    nodeSender.Create(nodeNum);

    // Install WifiPhys

    Ns2ExtWifiChannelHelper wifiChannel;
    wifiChannel.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");
    wifiChannel.AddPropagationLoss("ns3::FriisPropagationLossModel");

    Ns2ExtWifiPhyHelper wifiPhy = Ns2ExtWifiPhyHelper::Default();
    wifiPhy.SetChannel(wifiChannel.Create());
    wifiPhy.Set("UseConstantNoiseFloor", BooleanValue(true));
    wifiPhy.Set("ConstantNoiseFloor", DoubleValue(-99.0));
    wifiPhy.Set("PreambleCapture", BooleanValue(false));
    wifiPhy.Set("DataCapture", BooleanValue(false));

    WifiHelper wifi = WifiHelper::Default();
    wifi.SetMac("ns3::QosAdhocWifiMac",
                "EDCAParameterSet", StringValue(edcaParameterSet));

    wifi.SetRemoteStationManager("ns3::ConstantRateWifiManager",
                                 "DataMode", StringValue("wifia-6mbs"),
                                 "NonUnicastMode", StringValue("wifia-6mbs"));

    wifi.Install(wifiPhy, nodeListener);
    wifi.Install(wifiPhy, nodeSender);

    // Set up mobility

    MobilityHelper mobility;
    Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator>();
    for (int i = 0; i < nodeNum; ++i)
      {
        positionAlloc->Add(Vector(0.0, 0.0, 0.0));
      }
    mobility.SetPositionAllocator(positionAlloc);
    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");

    mobility.Install(nodeListener);
    mobility.Install(nodeSender);

    // Add packet socket handlers

    PacketSocketHelper packetSocket;
    packetSocket.Install(nodeListener);
    packetSocket.Install(nodeSender);

    // Configure packet socket: use broadcasts
    
    PacketSocketAddress socketBroadcast;
    socketBroadcast.SetAllDevices();
    socketBroadcast.SetPhysicalAddress(Mac48Address::GetBroadcast());
    socketBroadcast.SetProtocol(1);

    PacketSocketAddress socketToListener;

    if (unicast)
      {
        socketToListener.SetAllDevices();
        socketToListener.SetPhysicalAddress( nodeListener.Get(0)->GetDevice(0)->GetAddress() );
        socketToListener.SetProtocol(1);
      }
    else
      {
        socketToListener.SetAllDevices();
        socketToListener.SetPhysicalAddress( Mac48Address::GetBroadcast() );
        socketToListener.SetProtocol(1);
      }

    // Install three TrafficApplications on each node

    Ptr<QosTrafficPacketFactory> packetFactoryVO
      = CreateObject<QosTrafficPacketFactory>("AC", EnumValue(WifiQosTag::AC_VO),
                                              "Size", UintegerValue(200));

    Ptr<QosTrafficPacketFactory> packetFactoryVI
      = CreateObject<QosTrafficPacketFactory>("AC", EnumValue(WifiQosTag::AC_VI),
                                              "Size", UintegerValue(200));

    Ptr<QosTrafficPacketFactory> packetFactoryBE
      = CreateObject<QosTrafficPacketFactory>("AC", EnumValue(WifiQosTag::AC_BE),
                                              "Size", UintegerValue(200));

    Ptr<QosTrafficPacketFactory> packetFactoryBK
      = CreateObject<QosTrafficPacketFactory>("AC", EnumValue(WifiQosTag::AC_BK),
                                              "Size", UintegerValue(200));

    TrafficHelper trafficAppVO("ns3::PacketSocketFactory", socketToListener);
    trafficAppVO.SetAttribute("PacketFactory", PointerValue(packetFactoryVO) );
    trafficAppVO.SetAttribute("OnTime", RandomVariableValue( ConstantVariable(simulationTime) ));
    trafficAppVO.SetAttribute("OffTime", RandomVariableValue( UniformVariable(0.0,1.0/100.0) ));
    trafficAppVO.SetAttribute("Interval", RandomVariableValue( ExponentialVariable(1.0/100.0) ));

    ApplicationContainer appVO = trafficAppVO.Install(nodeSender);
    appVO.Start(Seconds(0.0));
    appVO.Stop(Seconds(simulationTime));

    TrafficHelper trafficAppVI("ns3::PacketSocketFactory", socketToListener);
    trafficAppVI.SetAttribute("PacketFactory", PointerValue(packetFactoryVI) );
    trafficAppVI.SetAttribute("OnTime", RandomVariableValue( ConstantVariable(simulationTime) ));
    trafficAppVI.SetAttribute("OffTime", RandomVariableValue( ConstantVariable(0.0) ));
    trafficAppVI.SetAttribute("Interval", RandomVariableValue( ExponentialVariable(1.0/100.0) ));

    ApplicationContainer appVI = trafficAppVI.Install(nodeSender);
    appVI.Start(Seconds(0.0));
    appVI.Stop(Seconds(simulationTime));

    TrafficHelper trafficAppBE("ns3::PacketSocketFactory", socketToListener);
    trafficAppBE.SetAttribute("PacketFactory", PointerValue(packetFactoryBE) );
    trafficAppBE.SetAttribute("OnTime", RandomVariableValue( ConstantVariable(simulationTime) ));
    trafficAppBE.SetAttribute("OffTime", RandomVariableValue( ConstantVariable(0.0) ));
    trafficAppBE.SetAttribute("Interval", RandomVariableValue( ExponentialVariable(1.0/100.0) ));

    ApplicationContainer appBE = trafficAppBE.Install(nodeSender);
    appBE.Start(Seconds(0.0));
    appBE.Stop(Seconds(simulationTime));

    TrafficHelper trafficAppBK("ns3::PacketSocketFactory", socketToListener);
    trafficAppBK.SetAttribute("PacketFactory", PointerValue(packetFactoryBK) );
    trafficAppBK.SetAttribute("OnTime", RandomVariableValue( ConstantVariable(simulationTime) ));
    trafficAppBK.SetAttribute("OffTime", RandomVariableValue( ConstantVariable(0.0) ));
    trafficAppBK.SetAttribute("Interval", RandomVariableValue( ExponentialVariable(1.0/100.0) ));

    ApplicationContainer appBK = trafficAppBK.Install(nodeSender);
    appBK.Start(Seconds(0.0));
    appBK.Stop(Seconds(simulationTime));

    // Install PacketSink on listener node

    PacketSinkHelper sink("ns3::PacketSocketFactory", socketBroadcast);

    ApplicationContainer app = sink.Install(nodeListener);
    app.Start(Seconds(0.0));
    app.Stop(Seconds(simulationTime));

    // connect Rx only on listening node.
    Config::Connect("/NodeList/0/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&Experiment::PacketSinkRxTrace, this));

    // Connect wifi traces
    
    //Config::Connect("/NodeList/*/DeviceList/*/Phy/State/Tx", MakeCallback(&Experiment::PhyTxTrace, this));
    //Config::Connect("/NodeList/*/DeviceList/*/Phy/State/RxOk", MakeCallback(&Experiment::PhyRxOkTrace, this));
    //Config::Connect("/NodeList/*/DeviceList/*/Phy/State/RxError", MakeCallback(&Experiment::PhyRxErrorTrace, this));

    //Config::Connect("/NodeList/*/DeviceList/*/Mac/MacTx", MakeCallback(&Experiment::DevTxTrace, this));

    //ConnectEventTrace();

    // Run simulation

    Simulator::Run();

    Simulator::Destroy();
  }
};

template <typename Container>
double calcMeanValue(const Container& c)
{
  return std::accumulate(c.begin(), c.end(), 0.0) / c.size();
}

template <typename Container>
double calcStandardDeviation(const Container& c)
{
  double squareSum = 0.0;
  double sum = 0.0;

  for (typename Container::const_iterator ei = c.begin();
       ei != c.end(); ++ei)
    {
      squareSum += (*ei) * (*ei);
      sum += *ei;
    }
  
  double mean = sum / c.size();
  return sqrt( (squareSum / c.size()) - (mean * mean) );
}

template <typename Container>
double calcConfidenceInterval(const Container& c)
{
  return 2.576 * calcStandardDeviation(c) / sqrt(c.size());
}

int main(int argc, char *argv[])
{
  // disable fragmentation
  Config::SetDefault("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue("2500"));
  Config::SetDefault("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue("2500"));

  LogComponentEnable("Main", LOG_LEVEL_ALL);

  CommandLine cmd;

  int fixedNumNodes = 0;
  int replications = 1;
  std::string edcaParameterSet = "802.11e";
  bool unicast = false;

  cmd.AddValue("EDCAParameterSet", "Select EDCA Parmeter set", edcaParameterSet);
  cmd.AddValue("Replications", "Perform independent replications", replications);
  cmd.AddValue("NumNodes", "Perform run for fixed number of nodes", fixedNumNodes);
  cmd.AddValue("Unicast", "Perform with ACKed unicast frames.", unicast);
  cmd.Parse(argc, argv);

  Gnuplot gnuplot("edca-traffic-streams.pdf");

  Gnuplot2dDataset::SetDefaultStyle(Gnuplot2dDataset::LINES_POINTS);
  Gnuplot2dDataset::SetDefaultErrorBars(Gnuplot2dDataset::X);
  Gnuplot2dDataset::SetDefaultExtra("pointsize 0.4");

  Gnuplot2dDataset datasetVO("VO");
  Gnuplot2dDataset datasetVI("VI");
  Gnuplot2dDataset datasetBE("BE");
  Gnuplot2dDataset datasetBK("BK");

  for(int nodeNum = 1; nodeNum <= 30; ++nodeNum)
    {
      if (fixedNumNodes != 0 && fixedNumNodes != nodeNum) continue;

      std::vector<double> throughputVO;
      std::vector<double> throughputVI;
      std::vector<double> throughputBE;
      std::vector<double> throughputBK;

      for(int rep = 0; rep < replications; ++rep)
        {
          SeedManager::SetRun(rep);

          Experiment experiment;
          experiment.RunOne(nodeNum, edcaParameterSet, unicast);
      
          std::cout << "# senders: " << nodeNum
                    << " throughputVO: " << std::fixed << experiment.GetThroughput(WifiQosTag::AC_VO)
                    << " throughputVI: " << std::fixed << experiment.GetThroughput(WifiQosTag::AC_VI)
                    << " throughputBE: " << std::fixed << experiment.GetThroughput(WifiQosTag::AC_BE)
                    << " throughputBK: " << std::fixed << experiment.GetThroughput(WifiQosTag::AC_BK)
                    << std::endl;

          throughputVO.push_back( experiment.GetThroughput(WifiQosTag::AC_VO) );
          throughputVI.push_back( experiment.GetThroughput(WifiQosTag::AC_VI) );
          throughputBE.push_back( experiment.GetThroughput(WifiQosTag::AC_BE) );
          throughputBK.push_back( experiment.GetThroughput(WifiQosTag::AC_BK) );
        }

      double meanVO = calcMeanValue(throughputVO);
      double confidenceVO = calcConfidenceInterval(throughputVO);

      double meanVI = calcMeanValue(throughputVI);
      double confidenceVI = calcConfidenceInterval(throughputVI);

      double meanBE = calcMeanValue(throughputBE);
      double confidenceBE = calcConfidenceInterval(throughputBE);

      double meanBK = calcMeanValue(throughputBK);
      double confidenceBK = calcConfidenceInterval(throughputBK);
      
      std::cout << "# nodes " << nodeNum
                << " rep " << replications
                << std::fixed
                << " throughputVO: " << meanVO << " bit/s " << confidenceVO
                << " throughputVI: " << meanVI << " bit/s " << confidenceVI
                << " throughputBE: " << meanBE << " bit/s " << confidenceBE
                << " throughputBK: " << meanBK << " bit/s " << confidenceBK
                << std::endl;

      datasetVO.Add(nodeNum, meanVO, confidenceVO);
      datasetVI.Add(nodeNum, meanVI, confidenceVI);
      datasetBE.Add(nodeNum, meanBE, confidenceBE);
      datasetBK.Add(nodeNum, meanBK, confidenceBK);
    }

  gnuplot.AddDataset(datasetVO);
  gnuplot.AddDataset(datasetVI);
  gnuplot.AddDataset(datasetBE);
  gnuplot.AddDataset(datasetBK);

  std::ofstream plotout("edca-traffic-streams.plot");
  gnuplot.GenerateOutput(plotout);

  return 0;
}
