/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

#include "ns3/propagation-loss-model.h"
#include "ns3/jakes-propagation-loss-model.h"

#include "ns3/constant-position-mobility-model.h"

#include "ns3/config.h"
#include "ns3/string.h"
#include "ns3/boolean.h"
#include "ns3/double.h"
#include "ns3/gnuplot.h"
#include "ns3/simulator.h"

#include <map>

using namespace ns3;

/// Round a double number to the given precision. e.g. dround(0.234, 0.1) = 0.2
/// and dround(0.257, 0.1) = 0.3
static double dround(double number, double precision)
{
  number /= precision;
  if (number >= 0)
    number = floor(number + 0.5);
  else
    number = ceil(number - 0.5);
  number *= precision;
  return number;
}

static Gnuplot2dDataset
TestProbabilistic (Ptr<PropagationLossModel> model, unsigned int distance = 100, unsigned int samples = 100000000)
{
  Ptr<ConstantPositionMobilityModel> a = CreateObject<ConstantPositionMobilityModel> ();
  Ptr<ConstantPositionMobilityModel> b = CreateObject<ConstantPositionMobilityModel> ();

  double txPowerDbm = +20; // dBm

  Gnuplot2dDataset dataset;

  typedef std::map<double, unsigned int> rxPowerMapType;

  // Take given number of samples from CalcRxPower() and show probability
  // density for discrete distances.
  {
    a->SetPosition (Vector (0.0, 0.0, 0.0));
    b->SetPosition (Vector (distance, 0.0, 0.0));

    rxPowerMapType rxPowerMap;

    for (unsigned int samp = 0; samp < samples; ++samp)
      {
        // CalcRxPower() returns dBm.
        double rxPowerDbm = model->CalcRxPower (txPowerDbm, a, b);
        rxPowerDbm = dround(rxPowerDbm, 0.1);

        rxPowerMap[ rxPowerDbm ] ++;

        Simulator::Stop (Seconds (0.01));
        Simulator::Run ();
      }

    for (rxPowerMapType::const_iterator i = rxPowerMap.begin();
         i != rxPowerMap.end(); ++i)
      {
        dataset.Add(i->first, (double)i->second / (double)samples);
      }
  }

  return dataset;
}

int main (int argc, char *argv[])
{
  Gnuplot plot("propagation-nakagami.pdf");
  plot.SetTitle("Nakagami-m Power Distributions");

  plot.AppendExtra("set xlabel 'rxPower (dBm)'");
  plot.AppendExtra("set ylabel 'Probability'");
  plot.AppendExtra("set key top left");
  plot.AppendExtra("set xrange[-20:]");

  //Gnuplot2dDataset::SetDefaultStyle(Gnuplot2dDataset::LINES_POINTS);
  //Gnuplot2dDataset::SetDefaultExtra("pointtype 3 pointsize 0.5");

  Gnuplot2dDataset::SetDefaultStyle(Gnuplot2dDataset::LINES);

  {
    Ptr<NakagamiPropagationLossModel> nak = CreateObject<NakagamiPropagationLossModel> ();
    nak->SetAttribute("m0", DoubleValue(0.75));
    nak->SetAttribute("m1", DoubleValue(0.75));
    nak->SetAttribute("m2", DoubleValue(0.75));

    Gnuplot2dDataset line = TestProbabilistic(nak);
    line.SetTitle("m=0.75");
    plot.AddDataset(line);
  }

  {
    Ptr<NakagamiPropagationLossModel> nak = CreateObject<NakagamiPropagationLossModel> ();
    nak->SetAttribute("m0", DoubleValue(1.0));
    nak->SetAttribute("m1", DoubleValue(1.0));
    nak->SetAttribute("m2", DoubleValue(1.0));

    Gnuplot2dDataset line = TestProbabilistic(nak);
    line.SetTitle("m=1.0");
    plot.AddDataset(line);
  }

  {
    Ptr<NakagamiPropagationLossModel> nak = CreateObject<NakagamiPropagationLossModel> ();
    nak->SetAttribute("m0", DoubleValue(2.0));
    nak->SetAttribute("m1", DoubleValue(2.0));
    nak->SetAttribute("m2", DoubleValue(2.0));

    Gnuplot2dDataset line = TestProbabilistic(nak);
    line.SetTitle("m=2.0");
    plot.AddDataset(line);
  }

  {
    Ptr<NakagamiPropagationLossModel> nak = CreateObject<NakagamiPropagationLossModel> ();
    nak->SetAttribute("m0", DoubleValue(3.0));
    nak->SetAttribute("m1", DoubleValue(3.0));
    nak->SetAttribute("m2", DoubleValue(3.0));

    Gnuplot2dDataset line = TestProbabilistic(nak);
    line.SetTitle("m=3.0");
    plot.AddDataset(line);
  }

  {
    Ptr<NakagamiPropagationLossModel> nak = CreateObject<NakagamiPropagationLossModel> ();
    nak->SetAttribute("m0", DoubleValue(5.0));
    nak->SetAttribute("m1", DoubleValue(5.0));
    nak->SetAttribute("m2", DoubleValue(5.0));

    Gnuplot2dDataset line = TestProbabilistic(nak);
    line.SetTitle("m=5.0");
    plot.AddDataset(line);
  }

  plot.GenerateOutput(std::cout);

  return 0;
}
