/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

#include "ns3/core-module.h"
#include "ns3/common-module.h"
#include "ns3/node-module.h"
#include "ns3/helper-module.h"
#include "ns3/contrib-module.h"
#include "ns3/wifi-module.h"
#include "ns3/mobility-module.h"
#include "ns3/traffic-application.h"

#include <iostream>
#include <fstream>
#include <map>

NS_LOG_COMPONENT_DEFINE("Main");

using namespace ns3;

#include "eventtrace.h"

class Experiment
{
private:

  static const double simulationTime = 10.0;

  static const double node1pos = 10;
  static const double node2pos = -10;

  uint32_t      m_packetsSent1;
  uint32_t      m_packetsTotalAt0;
  uint64_t      m_bytesTotalAt0;

  typedef std::map<uint64_t, Time> PacketQueueTimeType;
  std::map<uint64_t, Time> m_packetQueueTime;

public:
  double
  GetReceiveRatio() const
  {
    std::cerr << "sent: " << m_packetsSent1 << " recv " << m_packetsTotalAt0 << "\n";

    return (double)m_packetsTotalAt0 / (double)m_packetsSent1;
  }

  double
  GetThroughPut() const
  {
    // std::cerr << "sent: " << m_packetsSent1 << " recv " << m_packetsTotalAt0 << " = " << m_bytesTotalAt0 << " bytes\n";

    double throughPut = m_bytesTotalAt0 / simulationTime * 8;
    // std::cerr << "throughPut: " << throughPut << "\n";

    return throughPut;
  }

private:
  void
  PacketSinkRxTrace(std::string context, Ptr<const Packet> p, const Address &ad)
  {
    // std::cerr << context << " PacketSinkRx to=" << ad << " size=" << p->GetSize() << " p={" << *p << "}" << std::endl;

    m_packetsTotalAt0++;
    m_bytesTotalAt0 += p->GetSize();
  }

  void
  DevTxTrace(std::string context, Ptr<const Packet> p, Mac48Address address)
  {
    std::cerr << Simulator::Now() << " " << context << " DevTx to=" << address << " p={" << p->GetSize() << "}" << std::endl;

    m_packetsSent1++;
    m_packetQueueTime[p->GetUid()] = Simulator::Now();
  }

  void
  PhyTxTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag)
  {
    std::cerr << Simulator::Now() << " " << "PHYTX phytag={" << *phytag << "} size=" << p->GetSize() << " p={" << *p << "}" << std::endl;

    PacketQueueTimeType::const_iterator pi = m_packetQueueTime.find(p->GetUid());
    NS_ASSERT( pi != m_packetQueueTime.end() );

    std::cerr << "MAC access time: " << (Simulator::Now() - pi->second) << std::endl;
  }

  void
  PhyRxOkTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag)
  {
    std::cerr << "PHYRXOK phytag={" << *phytag << "} p={" << *p << "}" << std::endl;
  }

  void
  PhyRxErrorTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag, WifiPhy::RxErrorReason rxErrorReason)
  {
    std::cerr << "PHYRXERROR"
              << " reason=" << WifiPhy::RxErrorReasonToString(rxErrorReason)
              << " phytag={" << *phytag << "}"
              << " packet={" << *p << "}" << std::endl;
  }

public:

  
  Ptr<CloneTrafficPacketFactory> MakeQosTrafficPacketFactory(uint32_t packetsize, WifiQosTag::AccessCategory ac)
  {
    Ptr<Packet> qosPacket = Create<Packet>(packetsize);
    qosPacket->AddTag( WifiQosTag(ac) );

    Ptr<CloneTrafficPacketFactory> packetFactory = CreateObject<CloneTrafficPacketFactory>();
    packetFactory->SetPrototype(qosPacket);

    return packetFactory;
  }

  void
  RunOne(const char* wifiMode, uint32_t packetsize, int edcaAc)
  {
    m_packetsSent1 = 0;
    m_packetsTotalAt0 = 0;
    m_bytesTotalAt0 = 0;

    NodeContainer nodes;
    nodes.Create(3);

    // Install WifiPhy

    Ns2ExtWifiChannelHelper wifiChannel;
    wifiChannel.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");
    wifiChannel.AddPropagationLoss("ns3::FriisPropagationLossModel");

    Ns2ExtWifiPhyHelper wifiPhy = Ns2ExtWifiPhyHelper::Default();
    wifiPhy.SetChannel(wifiChannel.Create());
    wifiPhy.Set("UseConstantNoiseFloor", BooleanValue(true));
    wifiPhy.Set("ConstantNoiseFloor", DoubleValue(-99.0));

    WifiHelper wifi = WifiHelper::Default();
    wifi.SetMac("ns3::QosAdhocWifiMac");
    wifi.SetRemoteStationManager("ns3::ConstantRateWifiManager",
                                 "DataMode", StringValue("wifia-6mbs"));

    wifi.Install(wifiPhy, nodes);

    // Ns2ExtWifiPhyHelper::EnablePcapAll ("three-nodes-edca-accesstime");

    // Set up mobility

    MobilityHelper mobility;
    Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator>();
    positionAlloc->Add(Vector(0.0, 0.0, 0.0));
    positionAlloc->Add(Vector(node1pos, 0.0, 0.0));
    positionAlloc->Add(Vector(node2pos, 0.0, 0.0));
    mobility.SetPositionAllocator(positionAlloc);
    mobility.SetMobilityModel("ns3::StaticMobilityModel");
    mobility.Install(nodes);

    // Add packet socket handlers

    PacketSocketHelper packetSocket;
    packetSocket.Install(nodes);

    // Configure packet socket: receive everything

    PacketSocketAddress socket;
    socket.SetAllDevices();
    socket.SetPhysicalAddress(Mac48Address::GetBroadcast());
    socket.SetProtocol(1);

    // Install PacketSink on Node 0 (= Center)

    PacketSinkHelper sink("ns3::PacketSocketFactory", Address(socket));

    ApplicationContainer app = sink.Install(nodes.Get(0));
    app.Start(Seconds(0.0));
    app.Stop(Seconds(simulationTime));

    Config::Connect("/NodeList/0/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&Experiment::PacketSinkRxTrace, this));

#if 0
    // Install OnOffApplication on Node 1 send to 0


    OnOffHelper onoff("ns3::PacketSocketFactory", Address(socketTo0));
    onoff.SetAttribute("OnTime", RandomVariableValue( ConstantVariable(simulationTime) ));
    onoff.SetAttribute("OffTime", RandomVariableValue( ConstantVariable(0.0) ));
    onoff.SetAttribute("DataRate", DataRateValue(DataRate("100000kB/s")));
    onoff.SetAttribute("PacketSize", UintegerValue(400));

    app = onoff.Install(nodes.Get(1));
    app.Start(Seconds(0.0));
    app.Stop(Seconds(simulationTime));
#endif

    PacketSocketAddress socketTo0;
    socketTo0.SetAllDevices();
    socketTo0.SetPhysicalAddress( nodes.Get(0)->GetDevice(0)->GetAddress() );
    socketTo0.SetProtocol(1);

    Ptr<TrafficApplication> trafficApp1 = CreateObject<TrafficApplication>();
    trafficApp1->SetAttribute("PacketFactory", PointerValue( MakeQosTrafficPacketFactory(packetsize, WifiQosTag::AC_BK)) );
    trafficApp1->SetAttribute("OnTime", RandomVariableValue( ConstantVariable(simulationTime) ));
    trafficApp1->SetAttribute("OffTime", RandomVariableValue( ConstantVariable(0.0) ));
    trafficApp1->SetAttribute("Interval", RandomVariableValue( ExponentialVariable(0.001) ));
    //trafficApp1->SetAttribute("Interval", RandomVariableValue( ConstantVariable(1.0) ));
    trafficApp1->SetAttribute("Remote", AddressValue(socketTo0));
    trafficApp1->SetAttribute("Protocol", StringValue("ns3::PacketSocketFactory"));

    trafficApp1->SetNode(nodes.Get(1));
    trafficApp1->Start(Seconds(0.0));
    trafficApp1->Stop(Seconds(simulationTime));

    Ptr<TrafficApplication> trafficApp2 = CreateObject<TrafficApplication>();
    trafficApp2->SetAttribute("PacketFactory", PointerValue( MakeQosTrafficPacketFactory(packetsize, WifiQosTag::AC_VO)) );
    trafficApp2->SetAttribute("OnTime", RandomVariableValue( ConstantVariable(simulationTime) ));
    trafficApp2->SetAttribute("OffTime", RandomVariableValue( ConstantVariable(0.0) ));
    trafficApp2->SetAttribute("Interval", RandomVariableValue( ExponentialVariable(0.002) ));
    //trafficApp2->SetAttribute("Interval", RandomVariableValue( ConstantVariable(1.0) ));
    trafficApp2->SetAttribute("Remote", AddressValue(socketTo0));
    trafficApp2->SetAttribute("Protocol", StringValue("ns3::PacketSocketFactory"));

    trafficApp2->SetNode(nodes.Get(1));
    trafficApp2->Start(Seconds(0.0));
    trafficApp2->Stop(Seconds(simulationTime));

    // Connect wifi traces

    Config::Connect("/NodeList/1/DeviceList/*/Tx", MakeCallback(&Experiment::DevTxTrace, this));
    Config::Connect("/NodeList/1/DeviceList/*/Phy/State/Tx", MakeCallback(&Experiment::PhyTxTrace, this));

    // Config::Connect("/NodeList/*/DeviceList/*/Phy/State/RxOk", MakeCallback(&Experiment::PhyRxOkTrace, this));
    // Config::Connect("/NodeList/*/DeviceList/*/Phy/State/RxError", MakeCallback(&Experiment::PhyRxErrorTrace, this));

    //ConnectEventTrace();

    // Run simulation

    Simulator::Run();

    Simulator::Destroy();
    exit(1);
  }
};

int main(int argc, char *argv[])
{
  // disable fragmentation
  Config::SetDefault("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue("2500"));
  Config::SetDefault("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue("2500"));

  LogComponentEnable("Main", LOG_LEVEL_ALL);

  CommandLine cmd;
  cmd.Parse(argc, argv);

  NS_LOG_DEBUG("Starting");

  int edcaAcs[] = { -1, WifiQosTag::AC_VO, WifiQosTag::AC_VI, WifiQosTag::AC_BE, WifiQosTag::AC_BK };
  const char* wifiModes[] = { "wifia-6mbs", "wifia-24mbs", "wifia-54mbs" };
  unsigned int packetSizes[] = { 400, 2304 };

  for (unsigned int aci = 0; aci < sizeof(edcaAcs) / sizeof(edcaAcs[0]); ++aci)
    {
      for (unsigned int mi = 0; mi < sizeof(wifiModes) / sizeof(wifiModes[0]); ++mi)
        {
          for (unsigned int psi = 0; psi < sizeof(packetSizes) / sizeof(packetSizes[0]); ++psi)
            {
              Experiment experiment;
              experiment.RunOne(wifiModes[mi], packetSizes[psi], edcaAcs[aci]);

              std::cout << "Mode " << wifiModes[mi]
                        << " packetSize " << packetSizes[psi]
                        << " edcaAc " << edcaAcs[aci]
                        << " throughPuts: " << experiment.GetThroughPut() / 1e6
                        << std::endl;
            }
        }
    }

  return 0;
}
