/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

#include "ns3/core-module.h"
#include "ns3/common-module.h"
#include "ns3/node-module.h"
#include "ns3/helper-module.h"
#include "ns3/contrib-module.h"
#include "ns3/wifi-module.h"
#include "ns3/mobility-module.h"
#include "ns3/traffic-application.h"

#include <iostream>
#include <fstream>
#include <numeric>

NS_LOG_COMPONENT_DEFINE("Main");

using namespace ns3;

#include "eventtrace.h"

class Experiment
{
private:

  static const double simulationTime = 60.0;

  double        m_node1pos;

  uint32_t      m_packetsSent1;
  uint32_t      m_packetsTotalAt0;
  uint64_t      m_bytesTotalAt0;

  Time          m_idleTime;
  uint32_t      m_idleTimeCounter;

public:
  double
  GetReceiveRatio() const
  {
    std::cerr << "sent: " << m_packetsSent1 << " recv " << m_packetsTotalAt0 << "\n";

    return (double)m_packetsTotalAt0 / (double)m_packetsSent1;
  }

  double
  GetThroughPut() const
  {
    std::cout << "sent: " << m_packetsSent1 << " recv " << m_packetsTotalAt0 << " = " << m_bytesTotalAt0 << " bytes\n";

    double throughPut = 8 * m_bytesTotalAt0 / simulationTime;
    std::cout << std::fixed << "throughPut: " << throughPut << "\n";

    return throughPut;
  }

private:
  void
  PacketSinkRxTrace(std::string context, Ptr<const Packet> p, const Address &ad)
  {
    // std::cerr << context << " PacketSinkRx to=" << ad << " size=" << p->GetSize() << " p={" << *p << "}" << std::endl;

    m_packetsTotalAt0++;
    m_bytesTotalAt0 += p->GetSize();
  }

  void
  DevTxTrace(std::string context, Ptr<const Packet> p, Mac48Address address)
  {
    // std::cerr << context << " DevTx to=" << address << " p={" << p->GetSize() << "}" << std::endl;

    m_packetsSent1++;
  }

  void
  PhyTxTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag)
  {
    // std::cerr << "PHYTX phytag={" << *phytag << "} size=" << p->GetSize() << " p={" << *p << "}" << std::endl;

    m_packetsSent1++;
  }

  void
  PhyRxOkTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag)
  {
    std::cerr << "PHYRXOK phytag={" << *phytag << "} p={" << *p << "}" << std::endl;
  }

  void
  PhyRxErrorTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag, WifiPhy::RxErrorReason rxErrorReason)
  {
    std::cerr << "PHYRXERROR"
              << " reason=" << WifiPhy::RxErrorReasonToString(rxErrorReason)
              << " phytag={" << *phytag << "}"
              << " packet={" << *p << "}" << std::endl;
  }

  void
  Phy1StateTrace (std::string context, Time start, Time duration, enum WifiPhy::State state)
  {
    if (state != WifiPhy::IDLE) return;
/*
    std::cout << "EVENT event=wifistate context=" << context
              << " state=" << WifiPhy::StateToString(state)
              << " time=" << start
              << " duration=" << duration
              << std::endl;
*/
    m_idleTime += duration;
    m_idleTimeCounter++;
  }

public:

  void
  RunOne(const std::string& wifiMode, uint32_t packetsize, int edcaAc, bool withAck, const std::string& edcaParameterSet)
  {
    m_node1pos = 0;
    m_packetsSent1 = 0;
    m_packetsTotalAt0 = 0;
    m_bytesTotalAt0 = 0;

    m_idleTime = Seconds(0);
    m_idleTimeCounter = 0;

    NodeContainer nodes;
    nodes.Create(2);

    // Install WifiPhy

    Ns2ExtWifiChannelHelper wifiChannel;
    wifiChannel.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");
    wifiChannel.AddPropagationLoss("ns3::FriisPropagationLossModel");

    Ns2ExtWifiPhyHelper wifiPhy = Ns2ExtWifiPhyHelper::Default();
    wifiPhy.SetChannel(wifiChannel.Create());
    wifiPhy.Set("UseConstantNoiseFloor", BooleanValue(true));
    wifiPhy.Set("ConstantNoiseFloor", DoubleValue(-99.0));

    WifiHelper wifi = WifiHelper::Default();

    if (edcaAc < 0)
      wifi.SetMac("ns3::AdhocWifiMac");
    else
      wifi.SetMac("ns3::QosAdhocWifiMac",
                  "EDCAParameterSet", StringValue(edcaParameterSet));

    wifi.SetRemoteStationManager("ns3::ConstantRateWifiManager",
                                 "DataMode", StringValue(wifiMode),
                                 "NonUnicastMode", StringValue(wifiMode));

    wifi.Install(wifiPhy, nodes);

    // Ns2ExtWifiPhyHelper::EnablePcapAll ("two-nodes-edca-throughput");

    // Set up mobility

    MobilityHelper mobility;
    Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator>();
    positionAlloc->Add(Vector(0.0, 0.0, 0.0));
    positionAlloc->Add(Vector(m_node1pos, 0.0, 0.0));
    mobility.SetPositionAllocator(positionAlloc);
    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(nodes);

    // Add packet socket handlers

    PacketSocketHelper packetSocket;
    packetSocket.Install(nodes);

    // Configure packet socket: use broadcasts

    PacketSocketAddress socket;
    socket.SetAllDevices();
    socket.SetPhysicalAddress(Mac48Address::GetBroadcast());
    socket.SetProtocol(1);

    // Install PacketSink on Node 0 (= Center)

    PacketSinkHelper sink("ns3::PacketSocketFactory", Address(socket));

    ApplicationContainer app = sink.Install(nodes.Get(0));
    app.Start(Seconds(0.0));
    app.Stop(Seconds(simulationTime));

    Config::Connect("/NodeList/0/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&Experiment::PacketSinkRxTrace, this));

    // Install TrafficApplication on Node 1

    PacketSocketAddress socketTo0;
    socketTo0.SetAllDevices();
    socketTo0.SetProtocol(1);
    if (withAck)
      {
        socketTo0.SetPhysicalAddress( nodes.Get(0)->GetDevice(0)->GetAddress() );
      }
    else
      {
        socketTo0.SetPhysicalAddress( Mac48Address::GetBroadcast() );
      }

    Ptr<Packet> packet = Create<Packet>(packetsize);
    if (edcaAc >= 0)
      {
        packet->AddTag( WifiQosTag(WifiQosTag::AccessCategory(edcaAc)) );
      }

    Ptr<CloneTrafficPacketFactory> packetFactory = CreateObject<CloneTrafficPacketFactory>();
    packetFactory->SetPrototype(packet);

    Ptr<TrafficApplication> trafficApp = CreateObject<TrafficApplication>();
    trafficApp->SetAttribute("PacketFactory", PointerValue(packetFactory) );
    trafficApp->SetAttribute("OnTime", RandomVariableValue( ConstantVariable(simulationTime) ));
    trafficApp->SetAttribute("OffTime", RandomVariableValue( ConstantVariable(0.0) ));
    trafficApp->SetAttribute("Interval", RandomVariableValue( ConstantVariable(0.00005) ));
    trafficApp->SetAttribute("Remote", AddressValue(socketTo0));
    trafficApp->SetAttribute("Protocol", StringValue("ns3::PacketSocketFactory"));

    trafficApp->SetNode(nodes.Get(1));
    trafficApp->Start(Seconds(0.0));
    trafficApp->Stop(Seconds(simulationTime));

    // Connect wifi traces
    
    Config::Connect("/NodeList/1/DeviceList/*/Tx", MakeCallback(&Experiment::DevTxTrace, this));
    
    // Config::Connect("/NodeList/*/DeviceList/*/Phy/State/Tx", MakeCallback(&Experiment::PhyTxTrace, this));
    // Config::Connect("/NodeList/*/DeviceList/*/Phy/State/RxOk", MakeCallback(&Experiment::PhyRxOkTrace, this));
    // Config::Connect("/NodeList/*/DeviceList/*/Phy/State/RxError", MakeCallback(&Experiment::PhyRxErrorTrace, this));

    //ConnectEventTrace();

    Config::Connect ("/NodeList/1/DeviceList/*/Phy/State/State", MakeCallback (&Experiment::Phy1StateTrace, this));

    // Run simulation

    Simulator::Run();

    Simulator::Destroy();

    //std::cout << "idle average: " << m_idleTime / Scalar(m_idleTimeCounter) << "\n";
  }
};

template <typename Container>
double calcMeanValue(const Container& c)
{
  return std::accumulate(c.begin(), c.end(), 0.0) / c.size();
}

template <typename Container>
double calcStandardDeviation(const Container& c)
{
  double squareSum = 0.0;
  double sum = 0.0;

  for (typename Container::const_iterator ei = c.begin();
       ei != c.end(); ++ei)
    {
      squareSum += (*ei) * (*ei);
      sum += *ei;
    }
  
  double mean = sum / c.size();
  return sqrt( (squareSum / c.size()) - (mean * mean) );
}

int main(int argc, char *argv[])
{
  // disable fragmentation
  Config::SetDefault("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue("2500"));
  Config::SetDefault("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue("2500"));

  LogComponentEnable("Main", LOG_LEVEL_ALL);

  CommandLine cmd;

  int allModes = 0;
  std::string onlyMode;
  int replications = 1;
  std::string edcaParameterSet = "802.11e";

  cmd.AddValue("AllModes", "Test through all wifi modes", allModes);
  cmd.AddValue("OnlyMode", "Test through only this wifi mode", onlyMode);
  cmd.AddValue("EDCAParameterSet", "Select EDCA Parmeter set", edcaParameterSet);
  cmd.AddValue("Replications", "Perform independent replications", replications);
  cmd.Parse(argc, argv);

  const std::string allWifiModes[] = { "wifia-6mbs", "wifia-9mbs", "wifia-12mbs", "wifia-18mbs",
                                       "wifia-24mbs", "wifia-36mbs", "wifia-48mbs", "wifia-54mbs" };
  const std::string baseWifiModes[] = { "wifia-6mbs", "wifia-24mbs", "wifia-54mbs" };

  std::vector<std::string> wifiModes;

  if (onlyMode.size())
    {
      wifiModes.push_back(onlyMode);
    }
  else if (allModes)
    {
      wifiModes.assign(allWifiModes, allWifiModes + (sizeof(allWifiModes) / sizeof(allWifiModes[0])));
    }
  else
    {
      wifiModes.assign(baseWifiModes, baseWifiModes + (sizeof(baseWifiModes) / sizeof(baseWifiModes[0])));
    }

  int edcaAcs[] = { -1, WifiQosTag::AC_VO, WifiQosTag::AC_VI, WifiQosTag::AC_BE, WifiQosTag::AC_BK };
  unsigned int packetSizes[] = { 80, 200, 400, 2304 };

  for (unsigned int aci = 0; aci < sizeof(edcaAcs) / sizeof(edcaAcs[0]); ++aci)
    {
      for (unsigned int mi = 0; mi < wifiModes.size(); ++mi)
        {
          for (unsigned int psi = 0; psi < sizeof(packetSizes) / sizeof(packetSizes[0]); ++psi)
            {
              std::vector<double> throughPutNoAck;
              std::vector<double> throughPutAck;

              for(int rep = 0; rep < replications; ++rep)
                {
                  SeedManager::SetRun(rep);

                  Experiment experimentNoAck;
                  experimentNoAck.RunOne(wifiModes[mi], packetSizes[psi], edcaAcs[aci], false, edcaParameterSet);
                  
                  throughPutNoAck.push_back( experimentNoAck.GetThroughPut() );

                  Experiment experimentAck;
                  experimentAck.RunOne(wifiModes[mi], packetSizes[psi], edcaAcs[aci], true, edcaParameterSet);

                  throughPutAck.push_back( experimentAck.GetThroughPut() );
                }

              double meanNoAck = calcMeanValue(throughPutNoAck);
              double stddevNoAck = calcStandardDeviation(throughPutNoAck);
              double confidenceNoAck = 2.576 * stddevNoAck / sqrt(replications);

              double meanAck = calcMeanValue(throughPutAck);
              double stddevAck = calcStandardDeviation(throughPutAck);
              double confidenceAck = 2.576 * stddevAck / sqrt(replications);

              std::cout << "Mode " << wifiModes[mi]
                        << " packetSize " << packetSizes[psi]
                        << " edcaAc " << edcaAcs[aci]
                        << " rep " << replications
                        << std::fixed
                        << " throughPutNoAck: mean " << meanNoAck << " bit/s confidence " << confidenceNoAck
                        << " throughPutAck: mean " << meanAck << " bit/s confidence " << confidenceAck
                        << std::endl;
            }
        }
    }

  return 0;
}
