/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

#include "ns3/core-module.h"
#include "ns3/common-module.h"
#include "ns3/node-module.h"
#include "ns3/helper-module.h"
#include "ns3/wifi-module.h"
#include "ns3/mobility-module.h"
#include "ns3/contrib-module.h"

#include <iostream>
#include <fstream>
#include <numeric>

NS_LOG_COMPONENT_DEFINE("Main");

using namespace ns3;

class Experiment
{
private:

  double        m_node1pos;

  uint32_t      m_packetsSent1;
  uint32_t      m_packetsTotalAt0;

  double
  GetReceiveRatio() const
  {
    std::cerr << "sent: " << m_packetsSent1 << " recv " << m_packetsTotalAt0 << "\n";

    return (double)m_packetsTotalAt0 / (double)m_packetsSent1;
  }

  void
  PacketSinkRxTrace(std::string context, Ptr<const Packet> p, const Address &ad)
  {
    // std::cerr << context << " PacketSinkRx to=" << ad << " p={" << *p << "}" << std::endl;

    m_packetsTotalAt0++;
  }

  void
  DevTxTrace(std::string context, Ptr<const Packet> p, Mac48Address address)
  {
    // std::cerr << context << " DevTx to=" << address << " p={" << p->GetSize() << "}" << std::endl;

    m_packetsSent1++;
  }

  void
  PhyTxTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag)
  {
    std::cerr << "PHYTX phytag={" << *phytag << "} p={" << *p << "}" << std::endl;
  }

  void
  PhyRxOkTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag)
  {
    std::cerr << "PHYRXOK phytag={" << *phytag << "} p={" << *p << "}" << std::endl;
  }

  void
  PhyRxErrorTrace(std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag, WifiPhy::RxErrorReason rxErrorReason)
  {
    std::cerr << "PHYRXERROR"
              << " reason=" << WifiPhy::RxErrorReasonToString(rxErrorReason)
              << " phytag={" << *phytag << "}"
              << " packet={" << *p << "}" << std::endl;
  }

public:

  void
  RunOne(const WifiHelper &wifi, const WifiPhyHelper &wifiPhy, double node1pos)
  {
    const double simulationTime = 50.0;

    m_node1pos = node1pos;
    m_packetsSent1 = 0;
    m_packetsTotalAt0 = 0;

    NodeContainer nodes;
    nodes.Create(2);

    // Install WifiPhy and set up mobility
    wifi.Install(wifiPhy, nodes);

    MobilityHelper mobility;
    Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator>();
    positionAlloc->Add(Vector(0.0, 0.0, 0.0));
    positionAlloc->Add(Vector(node1pos, 0.0, 0.0));
    mobility.SetPositionAllocator(positionAlloc);
    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(nodes);

    // Add packet socket handlers

    PacketSocketHelper packetSocket;
    packetSocket.Install(nodes);

    // Configure packet socket: use broadcasts

    PacketSocketAddress socket;
    socket.SetAllDevices();
    socket.SetPhysicalAddress(Mac48Address::GetBroadcast());
    socket.SetProtocol(1);

    // Install PacketSink on Node 0 (= Center)

    PacketSinkHelper sink("ns3::PacketSocketFactory", Address(socket));

    ApplicationContainer app = sink.Install(nodes.Get(0));
    app.Start(Seconds(0.0));
    app.Stop(Seconds(simulationTime));

    Config::Connect("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&Experiment::PacketSinkRxTrace, this));

    // Install OnOffApplication on Node 1

    PacketSocketAddress socketTo0;
    socketTo0.SetAllDevices();
    // cannot use Unicast: packets are ACKed
    // socketTo0.SetPhysicalAddress( nodes.Get(0)->GetDevice(0)->GetAddress() );
    socketTo0.SetPhysicalAddress(Mac48Address::GetBroadcast());
    socketTo0.SetProtocol(1);

    OnOffHelper onoff("ns3::PacketSocketFactory", Address(socketTo0));
    onoff.SetAttribute("OnTime", RandomVariableValue( ConstantVariable(simulationTime) ));
    onoff.SetAttribute("OffTime", RandomVariableValue( ConstantVariable(0.0) ));
    onoff.SetAttribute("DataRate", DataRateValue(DataRate("160kB/s")));
    onoff.SetAttribute("PacketSize", UintegerValue(800));

    app = onoff.Install(nodes.Get(1));
    app.Start(Seconds(0.0));
    app.Stop(Seconds(simulationTime));

    // Connect wifi traces
    
    Config::Connect("/NodeList/1/DeviceList/*/Tx", MakeCallback(&Experiment::DevTxTrace, this));

    // Config::Connect("/NodeList/*/DeviceList/*/Phy/State/Tx", MakeCallback(&Experiment::PhyTxTrace, this));
    // Config::Connect("/NodeList/*/DeviceList/*/Phy/State/RxOk", MakeCallback(&Experiment::PhyRxOkTrace, this));
    // Config::Connect("/NodeList/*/DeviceList/*/Phy/State/RxError", MakeCallback(&Experiment::PhyRxErrorTrace, this));

    // Run simulation

    Simulator::Run();

    Simulator::Destroy();
  }

  template <typename WifiChannelHelper>
  Gnuplot2dDataset
  RunRange(const std::string& title, WifiChannelHelper& wifiChannel, const char* wifiMode, int replications);
};

double TestRangeMax(const std::string& wifiMode)
{
  return (wifiMode == "wifia-6mbs" ? 2500 :
          wifiMode == "wifia-9mbs" ? 2500 :
          wifiMode == "wifia-12mbs" ? 2000 :
          wifiMode == "wifia-18mbs" ? 2000 :
          wifiMode == "wifia-24mbs" ? 1000 :
          wifiMode == "wifia-36mbs" ? 1000 :
          wifiMode == "wifia-48mbs" ? 500 :
          wifiMode == "wifia-54mbs" ? 500 : 2500);
}

double TestRangeStep(const std::string& wifiMode)
{
  return (wifiMode == "wifia-6mbs" ? 10 :
          wifiMode == "wifia-9mbs" ? 10 :
          wifiMode == "wifia-12mbs" ? 10 :
          wifiMode == "wifia-18mbs" ? 10 :
          wifiMode == "wifia-24mbs" ? 5 :
          wifiMode == "wifia-36mbs" ? 5 :
          wifiMode == "wifia-48mbs" ? 2 :
          wifiMode == "wifia-54mbs" ? 2 : 10);
}

template <typename Container>
double calcMeanValue(const Container& c)
{
  return std::accumulate(c.begin(), c.end(), 0.0) / c.size();
}

template <typename Container>
double calcStandardDeviation(const Container& c)
{
  double squareSum = 0.0;
  double sum = 0.0;

  for (typename Container::const_iterator ei = c.begin();
       ei != c.end(); ++ei)
    {
      squareSum += (*ei) * (*ei);
      sum += *ei;
    }
  
  double mean = sum / c.size();
  return sqrt( (squareSum / c.size()) - (mean * mean) );
}

template <>
Gnuplot2dDataset
Experiment::RunRange(const std::string& title, YansWifiChannelHelper& wifiChannel, const char* wifiMode, int replications)
{
  wifiChannel.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");

  Gnuplot2dDataset dataset(title);

  for(double pos = 0.0; pos <= TestRangeMax(wifiMode); pos += TestRangeStep(wifiMode))
    {
      std::vector<double> receiveRatio;

      for(int r = 0; r < replications; ++r)
        {
          SeedManager::SetRun(r);

          std::cerr << title << " distance " << pos << " at " << wifiMode << " run " << r << "\n";

          YansWifiPhyHelper wifiPhy = YansWifiPhyHelper::Default();
          wifiPhy.SetChannel(wifiChannel.Create());

          WifiHelper wifi = WifiHelper::Default();
          wifi.SetMac("ns3::AdhocWifiMac");
          wifi.SetRemoteStationManager("ns3::ConstantRateWifiManager",
                                       "DataMode", StringValue(wifiMode),
                                       "NonUnicastMode", StringValue(wifiMode));

          RunOne(wifi, wifiPhy, pos);

          receiveRatio.push_back( GetReceiveRatio() );
        }

      if (receiveRatio.size() == 1)
        {
          dataset.Add(pos, GetReceiveRatio());
        }
      else
        {
          double mean = calcMeanValue(receiveRatio);
          double stddev = calcStandardDeviation(receiveRatio);

          std::cerr << "mean: " << mean << " stddev " << stddev << "\n";
          
          dataset.Add(pos, mean,
                      2.576 * stddev / sqrt(replications) );
        }
    }

  return dataset;
}

template <>
Gnuplot2dDataset
Experiment::RunRange(const std::string& title, Ns2ExtWifiChannelHelper& wifiChannel, const char* wifiMode, int replications)
{
  wifiChannel.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");

  Gnuplot2dDataset dataset(title);

  for(double pos = 0.0; pos <= TestRangeMax(wifiMode); pos += TestRangeStep(wifiMode))
    {
      std::vector<double> receiveRatio;

      for(int r = 0; r < replications; ++r)
        {
          SeedManager::SetRun(r);

          std::cerr << title << " distance " << pos << " at " << wifiMode << " run " << r << "\n";

          Ns2ExtWifiPhyHelper wifiPhy = Ns2ExtWifiPhyHelper::Default();
          wifiPhy.SetChannel(wifiChannel.Create());
          wifiPhy.Set("UseConstantNoiseFloor", BooleanValue(true));
          wifiPhy.Set("ConstantNoiseFloor", DoubleValue(-99.0));

          WifiHelper wifi = WifiHelper::Default();
          wifi.SetMac("ns3::AdhocWifiMac");
          wifi.SetRemoteStationManager("ns3::ConstantRateWifiManager",
                                       "DataMode", StringValue(wifiMode),
                                       "NonUnicastMode", StringValue(wifiMode));

          RunOne(wifi, wifiPhy, pos);

          receiveRatio.push_back( GetReceiveRatio() );
        }

      if (receiveRatio.size() == 1)
        {
          dataset.Add(pos, GetReceiveRatio());
        }
      else
        {
          double mean = calcMeanValue(receiveRatio);
          double stddev = calcStandardDeviation(receiveRatio);

          std::cerr << "mean: " << mean << " stddev " << stddev << "\n";
          
          dataset.Add(pos, mean,
                      2.576 * stddev / sqrt(replications) );
        }
    }

  return dataset;
}

template <typename WifiChannelHelper>
void RunExperiment(Gnuplot& gnuplot, const char* wifiMode, int allPropLoss, int replications)
{
  Experiment experiment;

  NS_LOG_DEBUG("Starting");

  // FriisPropagationLossModel
  {
    WifiChannelHelper wifiChannel;
    wifiChannel.AddPropagationLoss("ns3::FriisPropagationLossModel");

    GnuplotDataset dataset = experiment.RunRange("Friis", wifiChannel, wifiMode, 1);

    gnuplot.AddDataset(dataset);
  }

  if (allPropLoss)
    {
      // LogDistancePropagationLossModel
      {
        WifiChannelHelper wifiChannel;
        wifiChannel.AddPropagationLoss("ns3::LogDistancePropagationLossModel");

        GnuplotDataset dataset = experiment.RunRange("LogDistance (defaults)", wifiChannel, wifiMode, 1);

        gnuplot.AddDataset(dataset);
      }

      // LogDistancePropagationLossModel (exponent = 2.2)
      {
        WifiChannelHelper wifiChannel;
        wifiChannel.AddPropagationLoss("ns3::LogDistancePropagationLossModel",
                                       "Exponent", DoubleValue(2.2));

        GnuplotDataset dataset = experiment.RunRange("LogDistance (exponent = 2.2)", wifiChannel, wifiMode, 1);

        gnuplot.AddDataset(dataset);
      }

      // ThreeLogDistancePropagationLossModel
      {
        WifiChannelHelper wifiChannel;
        wifiChannel.AddPropagationLoss("ns3::ThreeLogDistancePropagationLossModel");

        GnuplotDataset dataset = experiment.RunRange("ThreeLogDistance (defaults)", wifiChannel, wifiMode, 1);

        gnuplot.AddDataset(dataset);
      }

      // ThreeLogDistancePropagationLossModel + NakagamiPropagationModel (default settings)
      {
        WifiChannelHelper wifiChannel;
        wifiChannel.AddPropagationLoss("ns3::ThreeLogDistancePropagationLossModel");
        wifiChannel.AddPropagationLoss("ns3::NakagamiPropagationLossModel");

        GnuplotDataset dataset = experiment.RunRange("ThreeLogDistance + Nakagami (defaults)", wifiChannel, wifiMode, replications);

        gnuplot.AddDataset(dataset);
      }

      // ThreeLogDistancePropagationLossModel + NakagamiPropagationModel (m = 1.0)
      {
        WifiChannelHelper wifiChannel;
        wifiChannel.AddPropagationLoss("ns3::ThreeLogDistancePropagationLossModel");
        wifiChannel.AddPropagationLoss("ns3::NakagamiPropagationLossModel",
                                       "m0", DoubleValue(1.0),
                                       "m1", DoubleValue(1.0),
                                       "m2", DoubleValue(1.0));

        GnuplotDataset dataset = experiment.RunRange("ThreeLogDistance + Nakagami (m = 1.0)", wifiChannel, wifiMode, replications);

        gnuplot.AddDataset(dataset);
      }

      // NakagamiPropagationModel (m = 3.0)
      {
        WifiChannelHelper wifiChannel;
        wifiChannel.AddPropagationLoss("ns3::ThreeLogDistancePropagationLossModel");
        wifiChannel.AddPropagationLoss("ns3::NakagamiPropagationLossModel",
                                       "m0", DoubleValue(3.0),
                                       "m1", DoubleValue(3.0),
                                       "m2", DoubleValue(3.0));

        GnuplotDataset dataset = experiment.RunRange("ThreeLogDistance + Nakagami (m = 3.0)", wifiChannel, wifiMode, replications);

        gnuplot.AddDataset(dataset);
      }

      // NakagamiPropagationModel (m = 5.0)
      {
        WifiChannelHelper wifiChannel;
        wifiChannel.AddPropagationLoss("ns3::ThreeLogDistancePropagationLossModel");
        wifiChannel.AddPropagationLoss("ns3::NakagamiPropagationLossModel",
                                       "m0", DoubleValue(5.0),
                                       "m1", DoubleValue(5.0),
                                       "m2", DoubleValue(5.0));

        GnuplotDataset dataset = experiment.RunRange("ThreeLogDistance + Nakagami (m = 5.0)", wifiChannel, wifiMode, replications);

        gnuplot.AddDataset(dataset);
      }
    }
}

int main(int argc, char *argv[])
{
  // disable fragmentation
  Config::SetDefault("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue("2200"));
  Config::SetDefault("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue("2200"));

  LogComponentEnable("Main", LOG_LEVEL_ALL);

  CommandLine cmd;
  int allModes = 0;
  int allPropLoss = 0;
  int allPhys = 0;
  int replications = 1;
  cmd.AddValue("AllModes", "Test through all wifi modes", allModes);
  cmd.AddValue("AllPropLoss", "Test through all propagation loss models", allPropLoss);
  cmd.AddValue("AllPhys", "Test through all WifiPhys", allPhys);
  cmd.AddValue("Replications", "Perform independent replications", replications);
  cmd.Parse(argc, argv);

  const char *allWifiModes[] = { "wifia-6mbs", "wifia-9mbs", "wifia-12mbs", "wifia-18mbs",
                                 "wifia-24mbs", "wifia-36mbs", "wifia-48mbs", "wifia-54mbs", NULL };
  const char *baseWifiModes[] = { "wifia-6mbs", NULL };

  const char** wifiModes = allModes ? allWifiModes : baseWifiModes;

  GnuplotCollection gnuplots("two-nodes-propagation.pdf");

  for (unsigned int mi = 0; wifiModes[mi]; ++mi)
    {
      Gnuplot gnuplot;
      gnuplot.SetTitle(std::string("Two Wireless Nodes with Ns2ExtWifiPhy at ") + wifiModes[mi]);
      gnuplot.AppendExtra("set xlabel 'Distance (m)'");
      gnuplot.AppendExtra("set ylabel 'Reception Probability'");

      Gnuplot2dDataset::SetDefaultStyle(Gnuplot2dDataset::LINES_POINTS);
      if (replications != 1)
        Gnuplot2dDataset::SetDefaultErrorBars(Gnuplot2dDataset::X);

      Gnuplot2dDataset::SetDefaultExtra("pointsize 0.4");

      RunExperiment<Ns2ExtWifiChannelHelper>(gnuplot, wifiModes[mi], allPropLoss, replications);

      gnuplots.AddPlot(gnuplot);
    }

  if (allPhys)
    {
      for (unsigned int mi = 0; wifiModes[mi]; ++mi)
        {
          Gnuplot gnuplot;
          gnuplot.SetTitle(std::string("Two Wireless Nodes with YansWifiPhy at ") + wifiModes[mi]);
          gnuplot.AppendExtra("set xlabel 'Distance (m)'");
          gnuplot.AppendExtra("set ylabel 'Reception Probability'");

          Gnuplot2dDataset::SetDefaultStyle(Gnuplot2dDataset::LINES_POINTS);
          if (replications != 1)
            Gnuplot2dDataset::SetDefaultErrorBars(Gnuplot2dDataset::X);

          Gnuplot2dDataset::SetDefaultExtra("pointsize 0.4");

          RunExperiment<YansWifiChannelHelper>(gnuplot, wifiModes[mi], allPropLoss, replications);

          gnuplots.AddPlot(gnuplot);
        }
    }

  gnuplots.GenerateOutput(std::cout);

  return 0;
}
