/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2005,2006 INRIA, 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Based on yans-wifi-phy.h by Mathieu Lacage <mathieu.lacage@sophia.inria.fr>
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

#ifndef NS2EXT_WIFI_PHY_H
#define NS2EXT_WIFI_PHY_H

#include <stdint.h>
#include "ns3/callback.h"
#include "ns3/event-id.h"
#include "ns3/packet.h"
#include "ns3/object.h"
#include "ns3/traced-callback.h"
#include "ns3/nstime.h"
#include "ns3/ptr.h"
#include "ns3/random-variable.h"
#include "wifi-phy.h"
#include "wifi-mode.h"
#include "wifi-preamble.h"
#include "wifi-phy-standard.h"
#include "interference-helper.h"


namespace ns3 {

class RandomUniform;
class RxEvent;
class Ns2ExtWifiChannel;
class YansWifiPhyStateHelper;


/**
 * \brief 802.11 PHY layer model compatible with ns-2's WirelessPhyExt.
 *
 * This class implements a ns3::WifiPhy but shows compatible behaviour with
 * ns-2's WirelessPhyExt class. The code is based on YansWifiPhy and was
 * modified to yield equal results.
 *
 * More information on the PHY layer modelling can be found at
 * http://dsn.tm.uni-karlsruhe.de/english/Overhaul_NS-2.php
 *
 * This PHY model depends on ns3::Ns2ExtWifiChannel to emulated channel
 * propagation loss and delay. See ns3::PropagationLossModel and
 * ns3::PropagationDelayModel classes for more information on reception ranges.
 */
class Ns2ExtWifiPhy : public WifiPhy
{
public:

  static TypeId GetTypeId (void);

  Ns2ExtWifiPhy ();
  virtual ~Ns2ExtWifiPhy ();

  void SetChannel (Ptr<Ns2ExtWifiChannel> channel);
  void StartReceivePacket (Ptr<Packet> packet, Ptr<WifiPhyTag> tag, double rxPowerDbm);

  void SetStandard (enum WifiPhyStandard standard);

  void SetRxNoiseFigure (double noiseFigureDb);
  double GetRxNoiseFigure (void) const;

  void SetUseConstantNoiseFloor (bool value);
  bool GetUseConstantNoiseFloor () const;

  void SetConstantNoiseFloor (double dbm);
  double GetConstantNoiseFloor () const;

  void SetTxPowerStart (double start);
  void SetTxPowerEnd (double end);
  void SetNTxPower (uint32_t n);
  void SetTxGain (double gain);
  void SetRxGain (double gain);
  void SetErrorRateModel (Ptr<ErrorRateModel> rate);
  void SetDevice (Ptr<Object> device);
  void SetMobility (Ptr<Object> mobility);

  double GetTxGain (void) const;
  double GetRxGain (void) const;
  Ptr<ErrorRateModel> GetErrorRateModel (void) const;
  Ptr<Object> GetDevice (void) const;
  Ptr<Object> GetMobility (void);

  /**
   * Set carrier sense threshold in dBm for ns-2 compatiblity mode.
   * \param dbm       in dBm (e.g. -94 dBm)
   */
  void SetCarrierSenseThreshold (double dbm);

  /**
   * \return carrier-sense threshold in dBm.
   */
  double GetCarrierSenseThreshold () const;

  virtual double GetTxPowerStart (void) const;
  virtual double GetTxPowerEnd (void) const;
  virtual uint32_t GetNTxPower (void) const;
  virtual void SetReceiveOkCallback (WifiPhy::RxOkCallback callback);
  virtual void SetReceiveErrorCallback (WifiPhy::RxErrorCallback callback);
  virtual void SendPacket (Ptr<const Packet> packet, WifiMode mode, enum WifiPreamble preamble, uint8_t txPowerLevel);
  virtual void RegisterListener (WifiPhyListener *listener);
  virtual bool IsStateCcaBusy (void);
  virtual bool IsStateIdle (void);
  virtual bool IsStateBusy (void);
  virtual bool IsStateSyncing (void);
  virtual bool IsStateRx (void);
  virtual bool IsStateTx (void);
  virtual Time GetStateDuration (void);
  virtual Time GetDelayUntilIdle (void);
  virtual Time GetLastRxStartTime (void) const;
  virtual Time CalculateTxDuration (uint32_t size, WifiMode payloadMode, enum WifiPreamble preamble) const;
  virtual uint32_t GetNModes (void) const;
  virtual WifiMode GetMode (uint32_t mode) const;
  virtual double CalculateSnr (WifiMode txMode, double ber) const;
  virtual Ptr<WifiChannel> GetChannel (void) const;

private:
  typedef std::vector<WifiMode> Modes;

private:
  Ns2ExtWifiPhy (const Ns2ExtWifiPhy &o);
  virtual void DoDispose (void);
  void Configure80211a (void);
  void ConfigureHolland (void);
  double GetEdThresholdW (void) const;
  double GetPowerDbm (uint8_t power) const;
  void EndReceivePacket (Ptr<Packet> packet, Ptr<WifiPhyTag> tag, Ptr<InterferenceHelper::Event> event);

private:
  double   m_txGainDb;
  double   m_rxGainDb;
  double   m_txPowerBaseDbm;
  double   m_txPowerEndDbm;
  uint32_t m_nTxPower;
  double   m_carrierSenseThresholdW;
  bool     m_doPreambleCapture;
  double   m_preambleCaptureThresholdDb;
  bool     m_doDataCapture;
  double   m_dataCaptureThresholdDb;

  Ptr<Ns2ExtWifiChannel> m_channel;
  Ptr<Object> m_device;
  Ptr<Object> m_mobility;
  Modes m_modes;
  WifiMode      m_headerMode;
  Time          m_preambleDuration;
  EventId m_endRxEvent;
  WifiPhyStandard m_standard;
  Ptr<YansWifiPhyStateHelper> m_state;
  InterferenceHelper m_interference;

  Ptr<Packet>      m_rxPacket;
  Ptr<WifiPhyTag>  m_rxPacketTag;
  Ptr<InterferenceHelper::Event> m_rxPacketEvent;
};

} // namespace ns3


#endif /* NS2EXT_WIFI_PHY_H */
