/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2005 INRIA, 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Based on adhoc-wifi-mac.cc by Mathieu Lacage <mathieu.lacage@sophia.inria.fr>
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

#include "qos-adhoc-wifi-mac.h"
#include "dca-txop.h"
#include "mac-low.h"
#include "mac-rx-middle.h"
#include "wifi-phy.h"
#include "dcf-manager.h"
#include "wifi-qos-tag.h"
#include "wifi-mac-trailer.h"
#include "ns3/pointer.h"
#include "ns3/packet.h"
#include "ns3/log.h"

NS_LOG_COMPONENT_DEFINE ("QosAdhocWifiMac");

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (QosAdhocWifiMac);

#undef NS_LOG_APPEND_CONTEXT
#define NS_LOG_APPEND_CONTEXT if (m_low != 0) {std::clog << "[mac=" << m_low->GetAddress () << "] ";}

TypeId 
QosAdhocWifiMac::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::QosAdhocWifiMac")
    .SetParent<WifiMac> ()
    .AddConstructor<QosAdhocWifiMac> ()
    .AddAttribute ("DcaTxopVO", "The DcaTxop object for AC_VO",
                   PointerValue (),
                   MakePointerAccessor (&QosAdhocWifiMac::DoGetDcaTxopVO),
                   MakePointerChecker<DcaTxop> ()) 
    .AddAttribute ("DcaTxopVI", "The DcaTxop object for AC_VI",
                   PointerValue (),
                   MakePointerAccessor (&QosAdhocWifiMac::DoGetDcaTxopVI),
                   MakePointerChecker<DcaTxop> ()) 
    .AddAttribute ("DcaTxopBE", "The DcaTxop object for AC_BE",
                   PointerValue (),
                   MakePointerAccessor (&QosAdhocWifiMac::DoGetDcaTxopBE),
                   MakePointerChecker<DcaTxop> ()) 
    .AddAttribute ("DcaTxopBK", "The DcaTxop object for AC_BK",
                   PointerValue (),
                   MakePointerAccessor (&QosAdhocWifiMac::DoGetDcaTxopBK),
                   MakePointerChecker<DcaTxop> ())
    .AddAttribute ("EDCAParameterSet", 
                   "Apply default EDCA parameters set from a standard.",
                   EnumValue (STANDARD_802_11p_D402),
                   MakeEnumAccessor (&QosAdhocWifiMac::SetDefaultEdcaParameters),
                   MakeEnumChecker (STANDARD_802_11e, "802.11e",
                                    STANDARD_802_11p_D402, "802.11p/D4.02"))
    ;
  return tid;
}

QosAdhocWifiMac::QosAdhocWifiMac ()
{
  m_rxMiddle = new MacRxMiddle ();
  m_rxMiddle->SetForwardCallback (MakeCallback (&QosAdhocWifiMac::ForwardUp, this));

  m_low = CreateObject<MacLow> ();
  m_low->SetRxCallback (MakeCallback (&MacRxMiddle::Receive, m_rxMiddle));
  m_low->SetBssid (GetBssid ());

  m_dcfManager = new DcfManager ();
  m_dcfManager->SetupLowListener (m_low);

  // create all four EDCAF - but leave DCF default parameters
  for (unsigned int aci = 0; aci < 4; ++aci)
    {
      m_edca[aci] = CreateObject<DcaTxop> ();
      m_edca[aci]->SetLow (m_low);
    }

  m_edca[3]->SetManager (m_dcfManager);
  m_edca[2]->SetManager (m_dcfManager);
  m_edca[0]->SetManager (m_dcfManager);
  m_edca[1]->SetManager (m_dcfManager);
}
QosAdhocWifiMac::~QosAdhocWifiMac ()
{}

void
QosAdhocWifiMac::DoDispose (void)
{
  delete m_rxMiddle;
  delete m_dcfManager;
  m_rxMiddle = 0;
  m_dcfManager = 0;
  m_low = 0;
  m_phy = 0;
  m_edca[0] = m_edca[1] = m_edca[2] = m_edca[3] = 0;
  m_stationManager = 0;
  WifiMac::DoDispose ();
}

void 
QosAdhocWifiMac::SetSlot (Time slotTime)
{
  m_dcfManager->SetSlot (slotTime);
  m_low->SetSlotTime (slotTime);
}
void 
QosAdhocWifiMac::SetSifs (Time sifs)
{
  m_dcfManager->SetSifs (sifs);
  m_low->SetSifs (sifs);
}
void 
QosAdhocWifiMac::SetEifsNoDifs (Time eifsNoDifs)
{
  m_dcfManager->SetEifsNoDifs (eifsNoDifs);
}
void 
QosAdhocWifiMac::SetAckTimeout (Time ackTimeout)
{
  m_low->SetAckTimeout (ackTimeout);
}
void 
QosAdhocWifiMac::SetCtsTimeout (Time ctsTimeout)
{
  m_low->SetCtsTimeout (ctsTimeout);
}
void 
QosAdhocWifiMac::SetPifs (Time pifs)
{
  m_low->SetPifs (pifs);
}
Time 
QosAdhocWifiMac::GetSlot (void) const
{
  return m_low->GetSlotTime ();
}
Time 
QosAdhocWifiMac::GetSifs (void) const
{
  return m_low->GetSifs ();
}
Time 
QosAdhocWifiMac::GetEifsNoDifs (void) const
{
  return m_dcfManager->GetEifsNoDifs();
}
Time 
QosAdhocWifiMac::GetAckTimeout (void) const
{
  return m_low->GetAckTimeout ();
}
Time 
QosAdhocWifiMac::GetCtsTimeout (void) const
{
  return m_low->GetCtsTimeout ();
}
Time 
QosAdhocWifiMac::GetPifs (void) const
{
  return m_low->GetPifs ();
}
void 
QosAdhocWifiMac::SetWifiPhy (Ptr<WifiPhy> phy)
{
  m_phy = phy;
  m_dcfManager->SetupPhyListener (phy);
  m_low->SetPhy (phy);
}
void 
QosAdhocWifiMac::SetWifiRemoteStationManager (Ptr<WifiRemoteStationManager> stationManager)
{
  m_stationManager = stationManager;
  m_edca[0]->SetWifiRemoteStationManager (stationManager);
  m_edca[1]->SetWifiRemoteStationManager (stationManager);
  m_edca[2]->SetWifiRemoteStationManager (stationManager);
  m_edca[3]->SetWifiRemoteStationManager (stationManager);
  m_low->SetWifiRemoteStationManager (stationManager);
}
void 
QosAdhocWifiMac::SetForwardUpCallback (Callback<void,Ptr<Packet>, Mac48Address, Mac48Address> upCallback)
{
  m_upCallback = upCallback;
}
void 
QosAdhocWifiMac::SetLinkUpCallback (Callback<void> linkUp)
{
  // an adhoc network is always UP.
  linkUp ();
}
void 
QosAdhocWifiMac::SetLinkDownCallback (Callback<void> linkDown)
{}
Mac48Address 
QosAdhocWifiMac::GetAddress (void) const
{
  return m_low->GetAddress ();
}
Ssid 
QosAdhocWifiMac::GetSsid (void) const
{
  return m_ssid;
}
Mac48Address 
QosAdhocWifiMac::GetBssid (void) const
{
  return m_low->GetBssid ();
}
void 
QosAdhocWifiMac::SetAddress (Mac48Address address)
{
  m_low->SetAddress (address);
  m_low->SetBssid (address);
  // XXX the bssid should be generated by the procedure
  // described in ieee802.11 section 11.1.3
}
void 
QosAdhocWifiMac::SetSsid (Ssid ssid)
{
  // XXX: here, we should start a special adhoc network
  m_ssid = ssid;
}

void
QosAdhocWifiMac::SetDefaultQosTag(const WifiQosTag& tag)
{
  m_defaultQosTag = tag;
}
WifiQosTag&
QosAdhocWifiMac::GetDefaultQosTag()
{
  return m_defaultQosTag;
}

void
QosAdhocWifiMac::SetDefaultEdcaParameters (enum DefaultEdcaParameterSet paramSet)
{
  NS_LOG_FUNCTION (this << paramSet);
  m_defaultParameterSet = paramSet;

  switch (paramSet) {
  case STANDARD_802_11e:
    {
      // 00 - AC_BE - Best Effort
      m_edca[0]->SetAifsn(3);
      m_edca[0]->SetCwMin(15);
      m_edca[0]->SetCwMax(1023);
      m_edca[0]->SetTxopLimit( MicroSeconds(0) );

      // 01 - AC_BK - Background
      m_edca[1]->SetAifsn(7);
      m_edca[1]->SetCwMin(15);
      m_edca[1]->SetCwMax(1023);
      m_edca[1]->SetTxopLimit( MicroSeconds(0) );

      // 10 - AC_VI - Video
      m_edca[2]->SetAifsn(2);
      m_edca[2]->SetCwMin(7);
      m_edca[2]->SetCwMax(15);
      m_edca[2]->SetTxopLimit( MicroSeconds(3008) );

      // 11 - AC_VO - Voice
      m_edca[3]->SetAifsn(2);
      m_edca[3]->SetCwMin(3);
      m_edca[3]->SetCwMax(7);
      m_edca[3]->SetTxopLimit( MicroSeconds(1504) );
    }
    break;
  case STANDARD_802_11p_D402:
    {
      // 00 - AC_BE - Best Effort
      m_edca[0]->SetAifsn(6);
      m_edca[0]->SetCwMin(7);
      m_edca[0]->SetCwMax(15);
      m_edca[0]->SetTxopLimit( MicroSeconds(0) );

      // 01 - AC_BK - Background
      m_edca[1]->SetAifsn(9);
      m_edca[1]->SetCwMin(15);
      m_edca[1]->SetCwMax(1023);
      m_edca[1]->SetTxopLimit( MicroSeconds(0) );

      // 10 - AC_VI - Video
      m_edca[2]->SetAifsn(3);
      m_edca[2]->SetCwMin(3);
      m_edca[2]->SetCwMax(7);
      m_edca[2]->SetTxopLimit( MicroSeconds(0) );

      // 11 - AC_VO - Voice
      m_edca[3]->SetAifsn(2);
      m_edca[3]->SetCwMin(3);
      m_edca[3]->SetCwMax(7);
      m_edca[3]->SetTxopLimit( MicroSeconds(0) );
    }
    break;
  default:
    NS_ASSERT (false);
    break;
  }
}

void 
QosAdhocWifiMac::Enqueue (Ptr<const Packet> packet, Mac48Address to, Mac48Address from)
{
  NS_FATAL_ERROR ("QosAdhoc does not support a from != m_low->GetAddress ()");
}
void 
QosAdhocWifiMac::Enqueue (Ptr<const Packet> packet, Mac48Address to)
{
  NS_LOG_FUNCTION (packet->GetSize () << to);

  // TODO: possibly use StationManager to figure out if we are allowed to send
  // QoS to the destination?

  WifiMacHeader hdr;
  hdr.SetType (WIFI_MAC_QOSDATA);
  hdr.SetAddr1 (to);
  hdr.SetAddr2 (m_low->GetAddress ());
  hdr.SetAddr3 (GetBssid ());
  hdr.SetDsNotFrom ();
  hdr.SetDsNotTo ();

  WifiRemoteStation *destination = m_stationManager->Lookup (to);
  if (destination->IsBrandNew ())
    {
      // in adhoc mode, we assume that every destination
      // supports all the rates we support.
      for (uint32_t i = 0; i < m_phy->GetNModes (); i++)
        {
          destination->AddSupportedMode (m_phy->GetMode (i));
        }
      destination->RecordDisassociated ();
    }

  // Determine QoS parameters

  WifiQosTag tag;
  if ( !packet->FindFirstMatchingTag (tag) )
    tag = m_defaultQosTag;

  uint32_t tid = tag.GetTid();
  NS_ASSERT(tid < 8);
  if (tid >= 8) tid = 0;
  hdr.SetQosTid(tid);

  hdr.SetQosEosp(0);
  hdr.SetQosTxopLimit(0);

  if (to.IsGroup())
    {
      hdr.SetQosAckPolicy(0);
    }
  else
    {
      hdr.SetQosAckPolicy( tag.GetAckPolicy() );
    }

  uint32_t ac = tag.GetAc();
  NS_ASSERT(ac < 4);

  m_edca[ac]->Queue (packet, hdr);

  NS_LOG_DEBUG("QosAdhoc::Enqueue ac=" << ac << " tid=" << tid);
}
bool 
QosAdhocWifiMac::SupportsSendFrom (void) const
{
  return false;
}

void 
QosAdhocWifiMac::ForwardUp (Ptr<Packet> packet, const WifiMacHeader *hdr)
{
  NS_LOG_DEBUG ("received size="<<packet->GetSize ()<<", from="<<hdr->GetAddr2 ());
  m_upCallback (packet, hdr->GetAddr2 (), hdr->GetAddr1 ());
}

Ptr<DcaTxop>
QosAdhocWifiMac::DoGetDcaTxopVO(void) const
{
  return m_edca[3];
}
Ptr<DcaTxop>
QosAdhocWifiMac::DoGetDcaTxopVI(void) const
{
  return m_edca[2];
}
Ptr<DcaTxop>
QosAdhocWifiMac::DoGetDcaTxopBE(void) const
{
  return m_edca[0];
}
Ptr<DcaTxop>
QosAdhocWifiMac::DoGetDcaTxopBK(void) const
{
  return m_edca[1];
}

} // namespace ns3
