/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2005 INRIA, 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Based on adhoc-wifi-mac.h by Mathieu Lacage <mathieu.lacage@sophia.inria.fr>
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

#ifndef QOS_ADHOC_WIFI_MAC_H
#define QOS_ADHOC_WIFI_MAC_H

#include "ns3/mac48-address.h"
#include "ns3/callback.h"
#include "ns3/packet.h"
#include "wifi-mac.h"
#include "wifi-qos-tag.h"

namespace ns3 {

class DcaTxop;
class Packet;
class WifiMacHeader;
class WifiPhy;
class DcfManager;
class MacLow;
class MacRxMiddle;

/**
 * \brief Adhoc Wifi MAC state machine supporting EDCA QoS.
 *
 * The Adhoc wifi mac does not associate or generate any other management
 * frames.
 */
class QosAdhocWifiMac : public WifiMac
{
public:
  enum DefaultEdcaParameterSet {
    STANDARD_802_11e,
    STANDARD_802_11p_D402,
  };

  typedef Callback<void, Ptr<Packet>, const Mac48Address &> ForwardCallback;

  static TypeId GetTypeId (void);

  QosAdhocWifiMac ();
  ~QosAdhocWifiMac ();

  // all inherited from WifiMac base class.
  virtual void SetSlot (Time slotTime);
  virtual void SetSifs (Time sifs);
  virtual void SetEifsNoDifs (Time eifsNoDifs);
  virtual void SetAckTimeout (Time ackTimeout);
  virtual void SetCtsTimeout (Time ctsTimeout);
  virtual void SetPifs (Time pifs);
  virtual Time GetSlot (void) const;
  virtual Time GetSifs (void) const;
  virtual Time GetEifsNoDifs (void) const;
  virtual Time GetAckTimeout (void) const;
  virtual Time GetCtsTimeout (void) const;
  virtual Time GetPifs (void) const;
  virtual void SetWifiPhy (Ptr<WifiPhy> phy);
  virtual void SetWifiRemoteStationManager (Ptr<WifiRemoteStationManager> stationManager);
  virtual void Enqueue (Ptr<const Packet> packet, Mac48Address to, Mac48Address from);
  virtual void Enqueue (Ptr<const Packet> packet, Mac48Address to);
  virtual bool SupportsSendFrom (void) const;
  virtual void SetForwardUpCallback (Callback<void,Ptr<Packet>, Mac48Address, Mac48Address> upCallback);
  virtual void SetLinkUpCallback (Callback<void> linkUp);
  virtual void SetLinkDownCallback (Callback<void> linkDown);
  virtual Mac48Address GetAddress (void) const;
  virtual Ssid GetSsid (void) const;
  virtual void SetAddress (Mac48Address address);
  virtual void SetSsid (Ssid ssid);
  virtual Mac48Address GetBssid (void) const;

  /**
   * Set default QoS parameters used for packets without attached WifiQosTag.
   */
  void          SetDefaultQosTag(const WifiQosTag& tag);

  /**
   * Return (writable) reference to default QoS parameters tag.
   */
  WifiQosTag&   GetDefaultQosTag();

  /**
   * Apply a set of default EDCA parameters from a standard.
   */
  void          SetDefaultEdcaParameters (enum DefaultEdcaParameterSet paramSet);

private:
  // inherited from Object base class.
  virtual void DoDispose (void);
  /* invoked by the MacLows. */
  void ForwardUp (Ptr<Packet> packet, const WifiMacHeader *hdr);
  QosAdhocWifiMac (const QosAdhocWifiMac & ctor_arg);
  QosAdhocWifiMac &operator = (const QosAdhocWifiMac &o);

  Ptr<DcaTxop> DoGetDcaTxopVO(void) const;
  Ptr<DcaTxop> DoGetDcaTxopVI(void) const;
  Ptr<DcaTxop> DoGetDcaTxopBE(void) const;
  Ptr<DcaTxop> DoGetDcaTxopBK(void) const;

  /// ECDA Functions and Queues: 0 = AC_BE, 1 = AC_BK, 2 = AC_VI, 3 = AC_VO
  Ptr<DcaTxop> m_edca[4];

  Callback<void, Ptr<Packet>, Mac48Address, Mac48Address> m_upCallback;
  Ptr<WifiRemoteStationManager> m_stationManager;
  Ptr<WifiPhy> m_phy;
  DcfManager *m_dcfManager;
  MacRxMiddle *m_rxMiddle;
  Ptr<MacLow> m_low;
  Ssid m_ssid;

  WifiQosTag    m_defaultQosTag;
  DefaultEdcaParameterSet m_defaultParameterSet;
};

} // namespace ns3

#endif /* QOS_ADHOC_WIFI_MAC_H */
