/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2005,2006 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Mathieu Lacage <mathieu.lacage@sophia.inria.fr>
 */

#include "wifi-phy.h"
#include "wifi-mode.h"
#include "wifi-channel.h"
#include "wifi-preamble.h"
#include "ns3/simulator.h"
#include "ns3/packet.h"
#include "ns3/random-variable.h"
#include "ns3/assert.h"
#include "ns3/log.h"
#include "ns3/double.h"
#include "ns3/integer.h"
#include "ns3/uinteger.h"
#include "ns3/enum.h"
#include "ns3/trace-source-accessor.h"
#include <math.h>

NS_LOG_COMPONENT_DEFINE ("WifiPhy");

namespace ns3 {

/****************************************************************
 *       This destructor is needed.
 ****************************************************************/

WifiPhyListener::~WifiPhyListener ()
{}

/***************************************************************
 *       WifiPhy Information Tagger
 ***************************************************************/ 

WifiPhyTag::WifiPhyTag ()
  : m_txed(false),
    m_rxed(false)
{}

void
WifiPhyTag::SetTxParameters(WifiPreamble wifiPreamble, WifiMode wifiMode, Time duration, uint8_t txPowerId, double txPowerDbm)
{
  m_txed = true;
  m_wifiPreamble = wifiPreamble;
  m_wifiMode = wifiMode;
  m_duration = duration;
  m_txPowerId = txPowerId;
  m_txPowerDbm = txPowerDbm;
} 
void
WifiPhyTag::SetRxParameters(double rxPowerDbm, double rxSnr, double rxPer)
{
  m_rxed = true;
  m_rxPowerDbm = rxPowerDbm;
  m_rxSnr = rxSnr;
  m_rxPer = rxPer;
} 

bool
WifiPhyTag::IsTxed() const
{
  return m_txed;
}
WifiPreamble
WifiPhyTag::GetWifiPreamble () const
{
  NS_ASSERT(m_txed);
  return m_wifiPreamble;
}
WifiMode
WifiPhyTag::GetWifiMode () const
{
  NS_ASSERT(m_txed);
  return m_wifiMode;
}
Time
WifiPhyTag::GetDuration () const
{
  NS_ASSERT(m_txed);
  return m_duration;
}
uint8_t
WifiPhyTag::GetTxPowerId () const
{
  NS_ASSERT(m_txed);
  return m_txPowerId;
}
double
WifiPhyTag::GetTxPowerDbm () const
{
  NS_ASSERT(m_txed);
  return m_txPowerDbm;
}
double
WifiPhyTag::GetTxPowerW() const
{
  NS_ASSERT(m_txed);
  return WifiPhy::DbmToW(m_txPowerDbm);
}

bool
WifiPhyTag::IsRxed() const
{
  return m_rxed;
}
double
WifiPhyTag::GetRxPowerDbm () const
{
  NS_ASSERT(m_rxed);
  return m_rxPowerDbm;
}
double
WifiPhyTag::GetRxPowerW () const
{
  NS_ASSERT(m_rxed);
  return WifiPhy::DbmToW(m_rxPowerDbm);
}
double
WifiPhyTag::GetRxSnr () const
{
  NS_ASSERT(m_rxed);
  return m_rxSnr;
}
double
WifiPhyTag::GetRxSnrDb () const
{
  NS_ASSERT(m_rxed);
  return WifiPhy::RatioToDb(m_rxSnr);
}
double
WifiPhyTag::GetRxPer () const
{
  NS_ASSERT(m_rxed);
  return m_rxPer;
}

TypeId 
WifiPhyTag::GetTypeId ()
{
  static TypeId tid = TypeId ("ns3::WifiPhyTag")
    .SetParent<Tag> ()
    .AddConstructor<WifiPhyTag> ()
    .AddAttribute ("WifiPreamble", 
                   "Packet wifi preamble type used in transmission.",
                   EmptyAttributeValue (),
                   MakeIntegerAccessor (&WifiPhyTag::GetWifiPreamble),
                   MakeIntegerChecker<int> ())
    .AddAttribute ("WifiMode", 
                   "Transmission mode of the packet.",
                   EmptyAttributeValue (),
                   MakeWifiModeAccessor (&WifiPhyTag::GetWifiMode),
                   MakeWifiModeChecker ())
    .AddAttribute ("Duration", 
                   "Duration of the packet in the aether.",
                   EmptyAttributeValue (),
                   MakeTimeAccessor (&WifiPhyTag::GetDuration),
                   MakeTimeChecker ())
    .AddAttribute ("TxPowerId", 
                   "Transmission power level id of the packet.",
                   EmptyAttributeValue (),
                   MakeUintegerAccessor (&WifiPhyTag::GetTxPowerId),
                   MakeUintegerChecker<uint8_t> ())
    .AddAttribute ("TxPowerDbm", 
                   "Transmission power level of the packet in dBm.",
                   EmptyAttributeValue (),
                   MakeDoubleAccessor (&WifiPhyTag::GetTxPowerDbm),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("TxPowerW", 
                   "Transmission power level of the packet in Watt.",
                   EmptyAttributeValue (),
                   MakeDoubleAccessor (&WifiPhyTag::GetTxPowerW),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("RxPowerDbm",
                   "Reception power of the packet in dBm.",
                   EmptyAttributeValue (),
                   MakeDoubleAccessor (&WifiPhyTag::GetRxPowerDbm),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("RxPowerW",
                   "Reception power of the packet in Watt.",
                   EmptyAttributeValue (),
                   MakeDoubleAccessor (&WifiPhyTag::GetRxPowerW),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("RxSnr", 
                   "Reception signal-to-noise-(and-interference)-ratio (SN(I)R) of the packet.",
                   EmptyAttributeValue (),
                   MakeDoubleAccessor (&WifiPhyTag::GetRxSnr),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("RxPer", 
                   "Reception packet-error-rate (PER) of the packet.",
                   EmptyAttributeValue (),
                   MakeDoubleAccessor (&WifiPhyTag::GetRxPer),
                   MakeDoubleChecker<double> ())
    ;
  return tid;
}
TypeId 
WifiPhyTag::GetInstanceTypeId () const
{
  return GetTypeId();
}
uint32_t 
WifiPhyTag::GetSerializedSize () const
{
  return sizeof(m_txed)
    + sizeof(m_wifiPreamble)
    + sizeof(m_wifiMode)
    + sizeof(m_duration)
    + sizeof(m_txPowerId)
    + sizeof(m_txPowerDbm)
    + sizeof(m_rxed)
    + sizeof(m_rxPowerDbm)
    + sizeof(m_rxSnr)
    + sizeof(m_rxPer)
    ;
}
void 
WifiPhyTag::Serialize (TagBuffer i) const
{
  i.Write ((uint8_t *)&m_txed, sizeof(m_txed));
  i.Write ((uint8_t *)&m_wifiPreamble, sizeof(m_wifiPreamble));
  i.Write ((uint8_t *)&m_wifiMode, sizeof(m_wifiMode));
  i.Write ((uint8_t *)&m_duration, sizeof(m_duration));
  i.Write ((uint8_t *)&m_txPowerId, sizeof(m_txPowerId));
  i.Write ((uint8_t *)&m_txPowerDbm, sizeof(m_txPowerDbm));
  i.Write ((uint8_t *)&m_rxed, sizeof(m_rxed));
  i.Write ((uint8_t *)&m_rxPowerDbm, sizeof(m_rxPowerDbm));
  i.Write ((uint8_t *)&m_rxSnr, sizeof(m_rxSnr));
  i.Write ((uint8_t *)&m_rxPer, sizeof(m_rxPer));
}
void 
WifiPhyTag::Deserialize (TagBuffer i)
{
  i.Read ((uint8_t *)&m_txed, sizeof(m_txed));
  i.Read ((uint8_t *)&m_wifiPreamble, sizeof(m_wifiPreamble));
  i.Read ((uint8_t *)&m_wifiMode, sizeof(m_wifiMode));
  i.Read ((uint8_t *)&m_duration, sizeof(m_duration));
  i.Read ((uint8_t *)&m_txPowerId, sizeof(m_txPowerId));
  i.Read ((uint8_t *)&m_txPowerDbm, sizeof(m_txPowerDbm));
  i.Read ((uint8_t *)&m_rxed, sizeof(m_rxed));
  i.Read ((uint8_t *)&m_rxPowerDbm, sizeof(m_rxPowerDbm));
  i.Read ((uint8_t *)&m_rxSnr, sizeof(m_rxSnr));
  i.Read ((uint8_t *)&m_rxPer, sizeof(m_rxPer));
}
void 
WifiPhyTag::Print (std::ostream &os) const
{
  os << "txed=" << m_txed;
  if (m_txed)
    {
      os << " wifiPreamble=" << m_wifiPreamble
         << " wifiMode=" << m_wifiMode
         << " duration=" << m_duration
         << " txPowerId=" << int(m_txPowerId)
         << " txPowerDbm=" << m_txPowerDbm;
    }

  os << " rxed=" << m_rxed;
  if (m_rxed)
    {
      os << " rxPowerDbm=" << m_rxPowerDbm
         << " rxSnr=" << m_rxSnr
         << " rxPer=" << m_rxPer;
    }
}

std::ostream& operator<< (std::ostream& os, const WifiPhyTag& tag)
{
  tag.Print (os);
  return os;
}

/****************************************************************
 *       The actual WifiPhy class
 ****************************************************************/

NS_OBJECT_ENSURE_REGISTERED (WifiPhy);

TypeId 
WifiPhy::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::WifiPhy")
    .SetParent<Object> ()
    .AddTraceSource ("PhyTxBegin", 
                     "Trace source indicating a packet has begun transmitting over the channel medium",
                     MakeTraceSourceAccessor (&WifiPhy::m_phyTxBeginTrace))
    .AddTraceSource ("PhyTxEnd", 
                     "Trace source indicating a packet has been completely transmitted over the channel",
                     MakeTraceSourceAccessor (&WifiPhy::m_phyTxEndTrace))
    .AddTraceSource ("PhyTxDrop", 
                     "Trace source indicating a packet has been dropped by the device during transmission",
                     MakeTraceSourceAccessor (&WifiPhy::m_phyTxDropTrace))
    .AddTraceSource ("PhyRxBegin", 
                     "Trace source indicating a packet has begun being received from the channel medium by the device",
                     MakeTraceSourceAccessor (&WifiPhy::m_phyRxBeginTrace))
    .AddTraceSource ("PhyRxEnd", 
                     "Trace source indicating a packet has been completely received from the channel medium by the device",
                     MakeTraceSourceAccessor (&WifiPhy::m_phyRxEndTrace))
    .AddTraceSource ("PhyRxDrop", 
                     "Trace source indicating a packet has been dropped by the device during reception",
                     MakeTraceSourceAccessor (&WifiPhy::m_phyRxDropTrace))
    .AddTraceSource ("PromiscSniffer", 
                     "Trace source simulating a promiscuous packet sniffer attached to the device",
                     MakeTraceSourceAccessor (&WifiPhy::m_phyPromiscSnifferTrace))
    ;
  return tid;
}

WifiPhy::WifiPhy ()
{
  NS_LOG_FUNCTION (this);
}

WifiPhy::~WifiPhy ()
{
  NS_LOG_FUNCTION (this);
}

WifiMode 
WifiPhy::Get6mba (void)
{
  static WifiMode mode = WifiModeFactory::CreateBpsk ("wifia-6mbs",
                                                      true,
                                                      20000000, 6000000, 12000000);
  return mode;
}
WifiMode 
WifiPhy::Get9mba (void)
{
  static WifiMode mode = WifiModeFactory::CreateBpsk ("wifia-9mbs",
                                                      false,
                                                      20000000, 9000000, 12000000);
  return mode;
}
WifiMode 
WifiPhy::Get12mba (void)
{
  static WifiMode mode = WifiModeFactory::CreateBpsk ("wifia-12mbs",
                                                      true,
                                                      20000000, 12000000, 24000000);
  return mode;
}
WifiMode 
WifiPhy::Get18mba (void)
{
  static WifiMode mode = WifiModeFactory::CreateBpsk ("wifia-18mbs",
                                                      false,
                                                      20000000, 18000000, 24000000);
  return mode;
}
WifiMode 
WifiPhy::Get24mba (void)
{
  static WifiMode mode = WifiModeFactory::CreateBpsk ("wifia-24mbs",
                                                      true,
                                                      20000000, 24000000, 48000000);
  return mode;
}
WifiMode 
WifiPhy::Get36mba (void)
{
  static WifiMode mode = WifiModeFactory::CreateBpsk ("wifia-36mbs",
                                                      false,
                                                      20000000, 36000000, 48000000);
  return mode;
}

WifiMode 
WifiPhy::Get48mba (void)
{
  static WifiMode mode = WifiModeFactory::CreateBpsk ("wifia-48mbs",
                                                      false,
                                                      20000000, 48000000, 72000000);
  return mode;
}

WifiMode 
WifiPhy::Get54mba (void)
{
  static WifiMode mode = WifiModeFactory::CreateBpsk ("wifia-54mbs",
                                                      false,
                                                      20000000, 54000000, 72000000);
  return mode;
}

const char*
WifiPhy::StateToString (enum State state)
{
  switch (state) {
  case IDLE:
    return "IDLE";
  case CCA_BUSY:
    return "CCA_BUSY";
  case TX:
    return "TX";
  case SYNCING:
    return "SYNCING";
  case RX:
    return "RX";
  default: // quiet compiler
    NS_ASSERT (false);
  }
  return "INVALID";
}

const char*
WifiPhy::RxErrorReasonToString (enum RxErrorReason rxErrorReason)
{
  switch(rxErrorReason)
    {
    case RXERROR_BAD_SIGNAL:
      return "RXERROR_BAD_SIGNAL";
    case RXERROR_ALREADY_RXING:
      return "RXERROR_ALREADY_RXING";
    case RXERROR_IN_TX:
      return "RXERROR_IN_TX";
    case RXERROR_TX_OVERRIDE:
      return "RXERROR_TX_OVERRIDE";
    case RXERROR_LOW_SIGNAL:
      return "RXERROR_LOW_SIGNAL";
    case RXERROR_CAPTURE_PREAMBLE:
      return "RXERROR_CAPTURE_PREAMBLE";
    case RXERROR_CAPTURE_DATA:
      return "RXERROR_CAPTURE_DATA";
    default: // quiet compiler
      NS_ASSERT (false);
    }
  return "INVALID";
}

void 
WifiPhy::NotifyTxBegin (Ptr<const Packet> packet)
{
  m_phyTxBeginTrace (packet);
}

void 
WifiPhy::NotifyTxEnd (Ptr<const Packet> packet)
{
  m_phyTxEndTrace (packet);
}

void 
WifiPhy::NotifyTxDrop (Ptr<const Packet> packet) 
{
  m_phyTxDropTrace (packet);
}

void 
WifiPhy::NotifyRxBegin (Ptr<const Packet> packet) 
{
  m_phyRxBeginTrace (packet);
}

void 
WifiPhy::NotifyRxEnd (Ptr<const Packet> packet) 
{
  m_phyRxEndTrace (packet);
}

void 
WifiPhy::NotifyRxDrop (Ptr<const Packet> packet) 
{
  m_phyRxDropTrace (packet);
}

void 
WifiPhy::NotifyPromiscSniff (Ptr<const Packet> packet) 
{
  m_phyPromiscSnifferTrace (packet);
}

double 
WifiPhy::DbmToW (double dBm)
{
  double mW = pow(10.0, dBm / 10.0);
  return mW / 1000.0;
}

double
WifiPhy::WToDbm (double watt)
{
  return 10.0 * log10(watt * 1000.0);
}

double 
WifiPhy::DbToRatio (double dB)
{
  double ratio = pow(10.0, dB / 10.0);
  return ratio;
}

double
WifiPhy::RatioToDb (double ratio)
{
  return 10.0 * log10(ratio);
}

} // namespace ns3

namespace {

static class Constructor
{
public:
  Constructor () {
    ns3::WifiPhy::Get6mba ();
    ns3::WifiPhy::Get9mba ();
    ns3::WifiPhy::Get12mba ();
    ns3::WifiPhy::Get18mba ();
    ns3::WifiPhy::Get24mba ();
    ns3::WifiPhy::Get36mba ();
    ns3::WifiPhy::Get48mba ();
    ns3::WifiPhy::Get54mba ();
  }
} g_constructor;
}
