/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

#include "wifi-qos-tag.h"

#include "ns3/uinteger.h"
#include "ns3/enum.h"

namespace ns3 {

TypeId
WifiQosTag::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::WifiQosTag")
    .SetParent<Tag> ()
    .AddConstructor<WifiQosTag> ()
    .AddAttribute ("TID", "The traffic identifier for the packet.",
                   UintegerValue (0),
                   MakeUintegerAccessor (&WifiQosTag::SetTid,
                                         &WifiQosTag::GetTid),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("AckPolicy", "ACK Policy for the packet.",
                   EnumValue (NORMAL_ACK),
                   MakeEnumAccessor (&WifiQosTag::SetAckPolicy,
                                     &WifiQosTag::GetAckPolicy),
                   MakeEnumChecker (NORMAL_ACK, "NORMAL_ACK",
                                    NO_ACK, "NO_ACK",
                                    BLOCK_ACK, "BLOCK_ACK") )
    ;
  return tid;
}
TypeId 
WifiQosTag::GetInstanceTypeId () const
{
  return GetTypeId ();
}

uint32_t 
WifiQosTag::GetSerializedSize () const
{
  return 4;
}
void 
WifiQosTag::Serialize (TagBuffer buf) const
{
  buf.WriteU32 (m_tid);
}
void 
WifiQosTag::Deserialize (TagBuffer buf)
{
  m_tid = buf.ReadU32 ();
}
void 
WifiQosTag::Print (std::ostream &os) const
{
  os << "tid=" << m_tid;
}

WifiQosTag::WifiQosTag ()
  : Tag (),
    m_tid (0), m_ackPolicy(NORMAL_ACK)
{}

WifiQosTag::WifiQosTag (uint32_t tid)
  : Tag (),
    m_tid (tid), m_ackPolicy(NORMAL_ACK)
{}

WifiQosTag::WifiQosTag (AccessCategory ac)
  : Tag (),
    m_ackPolicy(NORMAL_ACK)
{
  SetAc(ac);
}

void
WifiQosTag::SetTid (uint32_t tid)
{
  NS_ASSERT_MSG(tid < 16, "TID for 802.11 must be in the range 0-15");
  m_tid = tid;
}
uint32_t
WifiQosTag::GetTid () const
{
  return m_tid;
}

void
WifiQosTag::SetAc(AccessCategory ac)
{
  switch(ac)
    {
    case AC_BE:
      SetTid(0);
      return;
    case AC_BK:
      SetTid(1);
      return;
    case AC_VI:
      SetTid(4);
      return;
    case AC_VO:
      SetTid(6);
      return;
    }
}

WifiQosTag::AccessCategory
WifiQosTag::GetAcByUp(uint32_t up)
{
  NS_ASSERT(up < 8);
  switch(up)
    {
    case 0: case 3:
      return AC_BE;
    case 1: case 2:
      return AC_BK;
    case 4: case 5:
      return AC_VI;
    case 6: case 7:
      return AC_VO;
    default:
      NS_ASSERT_MSG(0, "ACs are only defined UPs < 8.");
      return AC_BE;
    }
}

WifiQosTag::AccessCategory
WifiQosTag::GetAc() const
{
  if (m_tid < 8)
    {
      return GetAcByUp(m_tid);
    }
  else
    {   
      NS_ASSERT_MSG(0, "ACs for TIDs > 7 are defined by TSPEC");
      return AC_BE;
    }
}

void
WifiQosTag::SetAckPolicy(AckPolicy ackPolicy)
{
  m_ackPolicy = ackPolicy;
}

WifiQosTag::AckPolicy
WifiQosTag::GetAckPolicy() const
{
  return m_ackPolicy;
}

} // namespace ns3
