/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

#ifndef WIFI_QOS_TAG_H
#define WIFI_QOS_TAG_H

#include "ns3/tag.h"

namespace ns3 {

/**
 * This Tag is used by upper layers to define 802.11 QoS parameters. Currently
 * this class only supports basic TCs (traffic categories) for EDCA. For HCCA
 * custom TSs (traffic streams) should be added.
 */
class WifiQosTag : public Tag
{
public:
  static TypeId GetTypeId ();
  virtual TypeId GetInstanceTypeId () const;
  virtual uint32_t GetSerializedSize () const;
  virtual void Serialize (TagBuffer buf) const;
  virtual void Deserialize (TagBuffer buf);
  virtual void Print (std::ostream &os) const;

  WifiQosTag ();

  /**
   * Constructor and initialize with TID.
   */
  WifiQosTag (uint32_t tid);

  /**
   * Set TID (Traffic Identifier) - this is the base value. Ranges from 0-15,
   * with 0-7 for TCs and 8-15 for TSs.
   */
  void SetTid (uint32_t tid);

  /**
   * \return TID currently set.
   */
  uint32_t GetTid () const;

  /**
   * EDCA ACs
   */
  enum AccessCategory { AC_BE = 0, AC_BK = 1, AC_VI = 2, AC_VO = 3 };

  /**
   * Constructor and initialize with AC.
   */
  WifiQosTag (AccessCategory ac);

  /**
   * Set TID by AC (access category). Picks the lowest TID matching the AC.
   */
  void SetAc(AccessCategory ac);

  /**
   * Infer EDCA AC from given UP.
   */
  static AccessCategory GetAcByUp(uint32_t up);

  /**
   * Infer EDCA AC from current TID.
   */
  AccessCategory GetAc() const;

  /**
   * ACK (acknowledgement) policies of packet. Not all policies are valid for
   * all types of packets. See 802.11-2007 standard section 7.1.3.5.3.
   */
  enum AckPolicy { NORMAL_ACK = 0, NO_ACK = 1, BLOCK_ACK = 3 };

  /**
   * Set ACK (acknowledgement) policy of packet.
   */
  void SetAckPolicy(AckPolicy ackPolicy);

  /**
   * \return ACK policy previously set.
   */
  AckPolicy GetAckPolicy() const;

private:
  uint32_t      m_tid;
  AckPolicy     m_ackPolicy;
};

} // namespace ns3

#endif /* WIFI_QOS_TAG_H */
