/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2005,2006 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Mathieu Lacage <mathieu.lacage@sophia.inria.fr>
 */
#include "yans-wifi-phy-state-helper.h"
#include "ns3/log.h"
#include "ns3/simulator.h"
#include "ns3/trace-source-accessor.h"

NS_LOG_COMPONENT_DEFINE ("YansWifiPhyStateHelper");

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (YansWifiPhyStateHelper);

TypeId 
YansWifiPhyStateHelper::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::YansWifiPhyStateHelper")
    .SetParent<Object> ()
    .AddConstructor<YansWifiPhyStateHelper> ()
    .AddTraceSource ("State",
                     "The state of the PHY layer",
                     MakeTraceSourceAccessor (&YansWifiPhyStateHelper::m_stateLogger))
    .AddTraceSource ("RxStart",
                     "Reception of a packet has started.",
                     MakeTraceSourceAccessor (&YansWifiPhyStateHelper::m_rxStartTrace))
    .AddTraceSource ("RxOk",
                     "A packet has been received successfully.",
                     MakeTraceSourceAccessor (&YansWifiPhyStateHelper::m_rxOkTrace))
    .AddTraceSource ("RxError",
                     "A packet has been received unsuccessfully.",
                     MakeTraceSourceAccessor (&YansWifiPhyStateHelper::m_rxErrorTrace))
    .AddTraceSource ("Tx", "Packet transmission is starting.",
                     MakeTraceSourceAccessor (&YansWifiPhyStateHelper::m_txTrace))
    ;
  return tid;
}

YansWifiPhyStateHelper::YansWifiPhyStateHelper ()
  : m_rxing (false),
    m_endTx (Seconds (0)),
    m_endRx (Seconds (0)),
    m_endCcaBusy (Seconds (0)),
    m_startTx (Seconds (0)),
    m_startRx (Seconds (0)),
    m_startCcaBusy (Seconds (0)),
    m_previousStateChangeTime (Seconds (0))
{
  NS_LOG_FUNCTION (this);
}

void 
YansWifiPhyStateHelper::SetReceiveOkCallback (WifiPhy::RxOkCallback callback)
{
  m_rxOkCallback = callback;
}
void 
YansWifiPhyStateHelper::SetReceiveErrorCallback (WifiPhy::RxErrorCallback callback)
{
  m_rxErrorCallback = callback;
}
void 
YansWifiPhyStateHelper::RegisterListener (WifiPhyListener *listener)
{
  m_listeners.push_back (listener);
}

bool 
YansWifiPhyStateHelper::IsStateCcaBusy (void)
{
  return (GetState () == WifiPhy::CCA_BUSY);
}
bool 
YansWifiPhyStateHelper::IsStateIdle (void)
{
  return (GetState () == WifiPhy::IDLE);
}
bool 
YansWifiPhyStateHelper::IsStateBusy (void)
{
  return (GetState () != WifiPhy::IDLE);
}
bool 
YansWifiPhyStateHelper::IsStateSyncing (void)
{
  return (GetState () == WifiPhy::SYNCING);
}
bool 
YansWifiPhyStateHelper::IsStateRx (void)
{
  return (GetState () == WifiPhy::RX);
}
bool 
YansWifiPhyStateHelper::IsStateTx (void)
{
  return (GetState () == WifiPhy::TX);
}

Time
YansWifiPhyStateHelper::GetStateDuration (void)
{
  return Simulator::Now () - m_previousStateChangeTime;
}

Time
YansWifiPhyStateHelper::GetDelayUntilIdle (void)
{
  Time retval;

  switch (GetState ()) {
  case WifiPhy::RX:
    retval = m_endRx - Simulator::Now ();
    break;
  case WifiPhy::TX:
    retval = m_endTx - Simulator::Now ();
    break;
  case WifiPhy::CCA_BUSY:
    retval = m_endCcaBusy - Simulator::Now ();
    break;
  case WifiPhy::IDLE:
    retval = Seconds (0);
    break;
  default:
    NS_ASSERT (false);
    // NOTREACHED
    retval = Seconds (0);
    break;
  }
  retval = Max (retval, Seconds (0));
  return retval;
}

Time 
YansWifiPhyStateHelper::GetLastRxStartTime (void) const
{
  return m_startRx;
}

enum WifiPhy::State 
YansWifiPhyStateHelper::GetState (void)
{
  if (m_endTx > Simulator::Now ()) 
    {
      return WifiPhy::TX;
    } 
  else if (m_rxing) 
    {
      return WifiPhy::RX;
    } 
  else if (m_endCcaBusy > Simulator::Now ()) 
    {
      return WifiPhy::CCA_BUSY;
    } 
  else 
    {
      return WifiPhy::IDLE;
    }
}


void 
YansWifiPhyStateHelper::NotifyTxStart (Time duration)
{
  for (Listeners::const_iterator i = m_listeners.begin (); i != m_listeners.end (); i++) {
    (*i)->NotifyTxStart (duration);
  }
}
void 
YansWifiPhyStateHelper::NotifyRxStart (Time duration)
{
  for (Listeners::const_iterator i = m_listeners.begin (); i != m_listeners.end (); i++) {
    (*i)->NotifyRxStart (duration);
  }
}
void 
YansWifiPhyStateHelper::NotifyRxEndOk (void)
{
  for (Listeners::const_iterator i = m_listeners.begin (); i != m_listeners.end (); i++) {
    (*i)->NotifyRxEndOk ();
  }
}
void 
YansWifiPhyStateHelper::NotifyRxEndError (void)
{
  for (Listeners::const_iterator i = m_listeners.begin (); i != m_listeners.end (); i++) {
    (*i)->NotifyRxEndError ();
  }
}
void 
YansWifiPhyStateHelper::NotifyMaybeCcaBusyStart (Time duration)
{
  for (Listeners::const_iterator i = m_listeners.begin (); i != m_listeners.end (); i++) {
    (*i)->NotifyMaybeCcaBusyStart (duration);
  }
}

void
YansWifiPhyStateHelper::LogPreviousIdleAndCcaBusyStates (void)
{
  Time now = Simulator::Now ();
  Time idleStart = Max (m_endCcaBusy, m_endRx);
  idleStart = Max (idleStart, m_endTx);
  NS_ASSERT (idleStart <= now);
  if (m_endCcaBusy > m_endRx && 
      m_endCcaBusy > m_endTx) {
    Time ccaBusyStart = Max (m_endTx, m_endRx);
    ccaBusyStart = Max (ccaBusyStart, m_startCcaBusy);
    m_stateLogger (ccaBusyStart, idleStart - ccaBusyStart, WifiPhy::CCA_BUSY);
  }
  m_stateLogger (idleStart, now - idleStart, WifiPhy::IDLE);
}

void
YansWifiPhyStateHelper::SwitchToTx (Time txDuration, Ptr<const Packet> packet, Ptr<const WifiPhyTag> tag)
{
  m_txTrace (packet, tag);
  NotifyTxStart (txDuration);
  Time now = Simulator::Now ();
  switch (GetState ()) {
  case WifiPhy::RX:
    /* The packet which is being received as well
     * as its endRx event are cancelled by the caller.
     */
    m_rxing = false;
    m_stateLogger (m_startRx, now - m_startRx, WifiPhy::RX);
    m_endRx = now;
    break;
  case WifiPhy::CCA_BUSY: {
    Time ccaStart = Max (m_endRx, m_endTx);
    ccaStart = Max (ccaStart, m_startCcaBusy);
    m_stateLogger (ccaStart, now - ccaStart, WifiPhy::CCA_BUSY);
  } break;
  case WifiPhy::IDLE:
    LogPreviousIdleAndCcaBusyStates ();
    break;
  default:
    NS_ASSERT (false);
    break;
  }
  m_stateLogger (now, txDuration, WifiPhy::TX);
  m_previousStateChangeTime = now;
  m_endTx = now + txDuration;
  m_startTx = now;
}
void
YansWifiPhyStateHelper::SwitchToRx (Time rxDuration)
{
  NS_ASSERT (IsStateIdle () || IsStateCcaBusy ());
  NS_ASSERT (!m_rxing);
  NotifyRxStart (rxDuration);
  Time now = Simulator::Now ();
  switch (GetState ()) {
  case WifiPhy::IDLE:
    LogPreviousIdleAndCcaBusyStates ();
    break;
  case WifiPhy::CCA_BUSY: {
    Time ccaStart = Max (m_endRx, m_endTx);
    ccaStart = Max (ccaStart, m_startCcaBusy);
    m_stateLogger (ccaStart, now - ccaStart, WifiPhy::CCA_BUSY);
  } break;
  case WifiPhy::RX:
  case WifiPhy::TX:
  default:
    NS_ASSERT (false);
    break;
  }
  m_previousStateChangeTime = now;
  m_rxing = true;
  m_startRx = now;
  m_endRx = now + rxDuration;
  NS_ASSERT (IsStateRx ());
}
void 
YansWifiPhyStateHelper::SwitchFromRxEndOk (Ptr<Packet> packet, Ptr<const WifiPhyTag> tag)
{
  NS_ASSERT (IsStateRx ());
  NS_ASSERT (m_rxing);

  m_rxOkTrace (packet, tag);
  NotifyRxEndOk ();
  DoSwitchFromRx ();
  if (!m_rxOkCallback.IsNull ())
    {
      m_rxOkCallback (packet, tag);
    }
}
void 
YansWifiPhyStateHelper::SwitchFromRxEndError (Ptr<const Packet> packet, Ptr<const WifiPhyTag> tag)
{
  NS_ASSERT (IsStateRx ());
  NS_ASSERT (m_rxing);

  m_rxErrorTrace (packet, tag, WifiPhy::RXERROR_BAD_SIGNAL);
  NotifyRxEndError ();
  DoSwitchFromRx ();
  if (!m_rxErrorCallback.IsNull ())
    {
      m_rxErrorCallback (packet, tag);
    }
}
void
YansWifiPhyStateHelper::SwitchFromRxAbort()
{
  NS_ASSERT (IsStateRx ());
  NS_ASSERT (m_rxing);

  NotifyRxEndError ();
  m_endRx = Simulator::Now ();
  DoSwitchFromRx ();
  NS_ASSERT (!IsStateRx ());
}

void
YansWifiPhyStateHelper::DoSwitchFromRx (void)
{
  NS_ASSERT (IsStateRx ());
  NS_ASSERT (m_rxing);

  Time now = Simulator::Now ();
  m_stateLogger (m_startRx, now - m_startRx, WifiPhy::RX);
  m_previousStateChangeTime = now;
  m_rxing = false;

  NS_ASSERT (IsStateIdle () || IsStateCcaBusy ());
}
void
YansWifiPhyStateHelper::SwitchMaybeToCcaBusy (Time duration)
{
  NotifyMaybeCcaBusyStart (duration);
  Time now = Simulator::Now ();
  switch (GetState ()) {
  case WifiPhy::IDLE:
    LogPreviousIdleAndCcaBusyStates ();
    break;
  case WifiPhy::CCA_BUSY:
    break;
  case WifiPhy::RX:
    break;
  case WifiPhy::TX:
    break;
  default:
    NS_ASSERT (false);
    break;
  }
  m_startCcaBusy = now;
  m_endCcaBusy = std::max (m_endCcaBusy, now + duration);
}

void
YansWifiPhyStateHelper::CallRxStartTrace (Ptr<const Packet> packet, Ptr<const WifiPhyTag> tag)
{
  m_rxStartTrace(packet, tag);
}

void
YansWifiPhyStateHelper::CallRxErrorTrace (Ptr<const Packet> packet, Ptr<const WifiPhyTag> tag, WifiPhy::RxErrorReason rxErrorReason)
{
  m_rxErrorTrace(packet, tag, rxErrorReason);
}

} // namespace ns3
